from qiskit import QuantumCircuit

from .ibm_fez_fidelity_estimation import estimate_circuit_fidelity_ibm_fez
from .quantinuum_h2_fidelity_estimation import estimate_circuit_fidelity_quantinuum_h2


def estimate_circuit_fidelity(
    circuit: QuantumCircuit,
    device: str,
    token: str = "",
    verbose: bool = False,
) -> float:
    """Estimate the fidelity of a quantum circuit on a given device.

    :param circuit: The quantum circuit to estimate the fidelity of. Must be a valid
                    Qiskit ``QuantumCircuit`` object.
    :param device: The device to estimate the fidelity on. Can be ``"quantinuum.h2"`` or
                ``"ibm.fez"``.
    :param token: The token to use for the online mode. Only required for ``"ibm.fez"``.
    :param verbose: Whether to print verbose output.

    :returns: The estimated fidelity of the circuit as a float.
    :rtype: float

    :raises ValueError: If the device is invalid.
    """
    if device == "quantinuum.h2":
        return estimate_circuit_fidelity_quantinuum_h2(circuit, verbose)
    if device == "ibm.fez":
        return estimate_circuit_fidelity_ibm_fez(circuit, token, verbose)
    raise ValueError(f"Invalid device: {device}")



from __future__ import annotations

import logging
import shutil
from functools import partial
from pathlib import Path

logger = logging.getLogger(__name__)


def has_extension(src: str, ext: str) -> bool:
    actual_extension = Path(src).suffix
    return ext == actual_extension


def copy_if(src: str, dst: str, *, include_ext_patterns: tuple[str, ...]) -> None:
    if any(has_extension(src, ext) for ext in include_ext_patterns):
        logger.debug("Copying %s to %s", src, dst)
        shutil.copy2(src, dst)


def copy_files(
    *,
    src: str | Path,
    dst: str | Path,
    include_ext_patterns: tuple[str, ...],
    exclude_ext_patterns: tuple[str, ...] = (),
) -> None:
    if not isinstance(include_ext_patterns, tuple):
        raise TypeError(
            f"include_ext_patterns should be {tuple} not {type(include_ext_patterns)}"
        )
    shutil.copytree(
        str(src),
        str(dst),
        dirs_exist_ok=True,
        copy_function=partial(copy_if, include_ext_patterns=include_ext_patterns),
        ignore=shutil.ignore_patterns(
            "*venv*",
            "__pycache__",
            ".git",
            ".ipynb_checkpoints",
            ".DS_Store",
            ".*",
            *exclude_ext_patterns,
        ),
    )
    logger.info(
        "Copied %s (excluding %s) files from %s to %s",
        include_ext_patterns,
        exclude_ext_patterns,
        src,
        dst,
    )



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np
from qiskit.primitives import (
    BaseSamplerV2,
    PrimitiveJob,
    PrimitiveResult,
    SamplerPubResult,
)
from qiskit.primitives.containers.sampler_pub import SamplerPub

from .provider import BlueQubitBackendV2

if TYPE_CHECKING:
    from collections.abc import Iterable

    from qiskit.circuit import QuantumCircuit
    from qiskit.primitives.containers import SamplerPubLike
    from qiskit.result import Result


@dataclass
class BQDataBin:
    meas: Result


class BlueQubitSamplerV2(BaseSamplerV2):
    """
    Qiskit SamplerV2 interface to BlueQubit cloud computation.
    """

    def __init__(
        self,
        *,
        backend: BlueQubitBackendV2 | None = None,
        backend_options: dict | None = None,
        run_options: dict | None = None,
    ):
        """
        Args:
            backend: The backend to run the primitive on.
            options: The options to specify the BlueQubit device type.
                For now, if you specify ``seed_simulator`` for deterministic output,
        """
        self._backend_options = backend_options if backend_options is not None else {}
        if backend is None:
            backend = BlueQubitBackendV2(**self._backend_options)

        self._backend = backend
        self._run_options = run_options if run_options is not None else {}

    def run(
        self, pubs: Iterable[SamplerPubLike], *, shots: int | None = None
    ) -> PrimitiveJob[PrimitiveResult[SamplerPubResult]]:
        if shots is None:
            shots = self._options.default_shots
        coerced_pubs = [SamplerPub.coerce(pub, shots) for pub in pubs]
        job = PrimitiveJob(self._run, coerced_pubs)
        job._submit()  # noqa: SLF001
        return job

    def _run(self, pubs: list[SamplerPub]) -> PrimitiveResult[SamplerPubResult]:
        pub_dict = defaultdict(list)
        # consolidate pubs with the same number of shots
        for i, pub in enumerate(pubs):
            pub_dict[pub.shots].append(i)

        results = [None] * len(pubs)
        for shots, lst in pub_dict.items():
            # run pubs with the same number of shots at once
            pub_results = self._run_pubs([pubs[i] for i in lst], shots)
            # reconstruct the result of pubs
            for i, pub_result in zip(lst, pub_results, strict=True):
                results[i] = pub_result
        return PrimitiveResult(results)

    def _run_pubs(self, pubs: list[SamplerPub], shots: int) -> list[SamplerPubResult]:
        """Compute results for pubs that all require the same value of ``shots``."""
        # prepare circuits
        bound_circuits = [pub.parameter_values.bind_all(pub.circuit) for pub in pubs]
        flatten_circuits: list[QuantumCircuit] = []
        for circuits in bound_circuits:
            flatten_circuits.extend(np.ravel(circuits).tolist())

        # run circuits
        job = self._backend.run(flatten_circuits, shots=shots, **self._run_options)
        results = job.result()

        pub_result = [
            SamplerPubResult(BQDataBin(meas=results[i])) for i in range(len(results))
        ]
        return pub_result



from functools import wraps

from qiskit.providers import JobError, JobStatus
from qiskit.providers import JobV1 as Job
from qiskit.result import Result

from bluequbit import job_metadata_constants


def requires_submit(func):
    """
    Decorator to ensure that a submit has been performed before
    calling the method.

    Args:
        func (callable): test function to be decorated.

    Returns:
        callable: the decorated function.
    """

    @wraps(func)
    def _wrapper(self, *args, **kwargs):
        if self._backend_job is None:
            raise JobError("Job not submitted yet!. You have to .submit() first!")
        return func(self, *args, **kwargs)

    return _wrapper


class BlueQubitQiskitJob(Job):
    def __init__(
        self,
        backend,
        job_id: str,
        fn,
        circuits,
        run_options=None,
    ):
        super().__init__(backend, job_id)
        self._job_id = job_id
        self._backend = backend
        self._backend_job = None
        self._circuits = circuits
        self._fn = fn
        self._run_options = run_options

    def submit(self):
        """Submit the job to the backend for execution."""
        if self._backend_job is not None:
            raise JobError("BlueQubit job already submitted!")
        self._backend_job = self._fn(self._circuits, self._run_options)

    @requires_submit
    def result(self) -> Result:
        """Return the results of the job."""
        if self._backend.execution_mode != "local":
            return self._backend.bq_client.wait(self._backend_job)
        return self._backend_job

    @requires_submit
    def cancel(self):
        """Attempt to cancel the job."""
        if self._backend.execution_mode == "local":
            raise JobError("Cannot cancel job in local mode")
        return self._backend.bq_client.cancel(self._backend_job)

    @requires_submit
    def status(self) -> JobStatus:
        """Return the status of the job, among the values of ``JobStatus``."""
        if self._backend.execution_mode == "local":
            return JobStatus.DONE
        job_status = [
            r.run_status for r in self._backend.bq_client.get(self._backend_job)
        ]
        if all(
            s in job_metadata_constants.JOB_RESULTS_READY_STATES for s in job_status
        ):
            return JobStatus.DONE
        if any(s in job_metadata_constants.JOB_NON_TERMINAL_STATES for s in job_status):
            return JobStatus.RUNNING
        return JobStatus.ERROR



from qiskit import QuantumCircuit, transpile

# --- H2 Fidelity Constants (Typical) ---
# Based on "Typical" column in Table 1
INFIDELITY_1Q_TYPICAL = 3e-5
FIDELITY_1Q_TYPICAL = 1.0 - INFIDELITY_1Q_TYPICAL

INFIDELITY_2Q_TYPICAL = 1e-3
FIDELITY_2Q_TYPICAL = 1.0 - INFIDELITY_2Q_TYPICAL

INFIDELITY_SPAM_TYPICAL = 1e-3
FIDELITY_SPAM_TYPICAL = 1.0 - INFIDELITY_SPAM_TYPICAL

# --- H2 Fidelity Constants (Min) ---
# Based on "Max" column in Table 1
INFIDELITY_1Q_MAX = 2e-4
FIDELITY_1Q_MIN = 1.0 - INFIDELITY_1Q_MAX

INFIDELITY_2Q_MAX = 2e-3
FIDELITY_2Q_MIN = 1.0 - INFIDELITY_2Q_MAX

INFIDELITY_SPAM_MAX = 5e-3
FIDELITY_SPAM_MIN = 1.0 - INFIDELITY_SPAM_MAX


def calculate_circuit_fidelity_quantinuum_h2(
    circuit: QuantumCircuit, include_readout=True, mode: str = "typical"
) -> dict:
    """Calculate the fidelity estimation of a quantum circuit on the Quantinuum H2.

    This calculation is simple and assumes all-to-all connectivity. It does
    not transpile the circuit; it only counts the gates in the provided
    transpiled circuit. It does not account for memory errors.

    :param circuit: The transpiled quantum circuit to estimate the fidelity of.
    :param include_readout: Whether to include SPAM errors in the estimate.
    :param mode: The fidelity mode to use: ``"typical"`` (default) or ``"min"``.

    :returns: A dictionary with fidelity estimates and gate counts.
    :rtype: dict
    """
    num_1q_gates = 0
    num_2q_gates = 0
    num_measurements = 0

    # Analyze each instruction in the circuit
    for instruction_data in circuit.data:
        # Handle both old (tuple) and new (InstructionData) Qiskit API
        if hasattr(instruction_data, "operation"):
            instruction = instruction_data.operation
            qargs = instruction_data.qubits
        else:
            instruction, qargs, _ = instruction_data

        gate_name = instruction.name
        num_qubits = len(qargs)

        # Skip non-gate operations
        if gate_name in ["barrier", "delay"]:
            continue

        # Categorize operation
        if gate_name == "measure":
            num_measurements += 1
        elif num_qubits == 1:
            num_1q_gates += 1
        elif num_qubits == 2:
            num_2q_gates += 1
        # We can ignore 3+ qubit gates
        # as they would be decomposed during transpilation.

    # --- Select Fidelities Based on Mode ---
    if mode.lower() == "min":
        infidelity_1q = INFIDELITY_1Q_MAX
        fidelity_1q = FIDELITY_1Q_MIN
        infidelity_2q = INFIDELITY_2Q_MAX
        fidelity_2q = FIDELITY_2Q_MIN
        infidelity_spam = INFIDELITY_SPAM_MAX
        fidelity_spam = FIDELITY_SPAM_MIN
        mode_name = "Min"
    else:
        # Default to typical
        infidelity_1q = INFIDELITY_1Q_TYPICAL
        fidelity_1q = FIDELITY_1Q_TYPICAL
        infidelity_2q = INFIDELITY_2Q_TYPICAL
        fidelity_2q = FIDELITY_2Q_TYPICAL
        infidelity_spam = INFIDELITY_SPAM_TYPICAL
        fidelity_spam = FIDELITY_SPAM_TYPICAL
        mode_name = "Typical"

    # --- Calculate Fidelity ---

    # Method 1: Product of fidelities (more accurate)
    fidelity_product = 1.0
    fidelity_product *= fidelity_1q**num_1q_gates
    fidelity_product *= fidelity_2q**num_2q_gates
    if include_readout:
        # Apply SPAM fidelity for each measurement operation
        fidelity_product *= fidelity_spam**num_measurements

    # Method 2: Linear approximation (sum of errors)
    total_error = 0.0
    total_error += num_1q_gates * infidelity_1q
    total_error += num_2q_gates * infidelity_2q
    if include_readout:
        total_error += num_measurements * infidelity_spam

    fidelity_linear = 1.0 - total_error

    return {
        "fidelity_product": fidelity_product,
        "fidelity_linear": fidelity_linear,
        "total_error": total_error,
        "num_single_qubit_gates": num_1q_gates,
        "num_two_qubit_gates": num_2q_gates,
        "num_measurements": num_measurements,
        # Pass back info for the report
        "mode": mode_name,
        "infidelity_1q": infidelity_1q,
        "fidelity_1q": fidelity_1q,
        "infidelity_2q": infidelity_2q,
        "fidelity_2q": fidelity_2q,
        "infidelity_spam": infidelity_spam,
        "fidelity_spam": fidelity_spam,
    }


def get_fidelity_report_string_quantinuum_h2(
    fidelity_result: dict, circuit: QuantumCircuit, original_circuit: QuantumCircuit
) -> str:
    """Get a formatted report string of the H2 fidelity estimation.

    :param fidelity_result: Result dictionary from ``calculate_circuit_fidelity_quantinuum_h2``.
    :param circuit: Transpiled quantum circuit.
    :param original_circuit: Original quantum circuit.

    :returns: A formatted report string.
    :rtype: str
    """
    report_string = ""
    report_string += "=" * 70 + "\n"
    report_string += "QUANTINUUM H2 FIDELITY ESTIMATION REPORT\n"
    report_string += "=" * 70 + "\n"

    report_string += "\nOriginal Circuit:\n"
    report_string += f"  Gates (total): {len(original_circuit.data)}\n"
    report_string += f"  Depth:         {original_circuit.depth()}\n"
    report_string += f"  Qubits:        {original_circuit.num_qubits}\n"

    report_string += "\nTranspiled Circuit:\n"
    report_string += f"  Gates (total): {len(circuit.data)}\n"
    report_string += f"  Depth:         {circuit.depth()}\n"
    report_string += f"  Qubits:        {circuit.num_qubits}\n"

    # Get mode and fidelities from the result dict
    mode_name = fidelity_result.get("mode", "Typical")
    fidelity_1q = fidelity_result["fidelity_1q"]
    infidelity_1q = fidelity_result["infidelity_1q"]
    fidelity_2q = fidelity_result["fidelity_2q"]
    infidelity_2q = fidelity_result["infidelity_2q"]
    fidelity_spam = fidelity_result["fidelity_spam"]
    infidelity_spam = fidelity_result["infidelity_spam"]

    report_string += f"\nFidelity Estimates (based on {mode_name} fidelities):\n"
    report_string += (
        f"  Fidelity (product method): {fidelity_result['fidelity_product']:.6f}\n"
    )
    report_string += (
        f"  Fidelity (linear approx):  {fidelity_result['fidelity_linear']:.6f}\n"
    )
    report_string += f"  Total error sum:        {fidelity_result['total_error']:.6f}\n"

    report_string += "\nGate Counts (from transpiled circuit):\n"
    report_string += (
        f"  Single-qubit gates: {fidelity_result['num_single_qubit_gates']}\n"
    )
    report_string += f"  Two-qubit gates:    {fidelity_result['num_two_qubit_gates']}\n"
    report_string += f"  Measurements:       {fidelity_result['num_measurements']}\n"

    report_string += "\n" + "=" * 70 + "\n"
    report_string += "NOTES:\n"
    report_string += "=" * 70 + "\n"
    report_string += "• Assumes all-to-all connectivity (no connectivity constraints during transpilation). \n"
    report_string += (
        "• Transpiled to 'u' (general 1Q) and 'cx' (general 2Q proxy) gates.\n"
    )
    report_string += f"• Uses '{mode_name}' fidelities from H2 data sheet: \n"
    report_string += (
        f"  - 1Q Fidelity: {fidelity_1q:.6f} (Error: {infidelity_1q:.1e})\n"
    )
    report_string += (
        f"  - 2Q Fidelity: {fidelity_2q:.6f} (Error: {infidelity_2q:.1e})\n"
    )
    report_string += (
        f"  - SPAM Fidelity: {fidelity_spam:.6f} (Error: {infidelity_spam:.1e})\n"
    )
    report_string += "• This estimate does not account for memory errors. \n"
    report_string += "• This estimate does not account for crosstalk between qubits. \n"
    return report_string


def estimate_circuit_fidelity_quantinuum_h2(
    circuit: QuantumCircuit, verbose: bool = False, mode: str = "typical"
) -> float:
    """Estimate the fidelity of a quantum circuit on the Quantinuum H2.

    :param circuit: The quantum circuit to estimate the fidelity of.
    :param verbose: Whether to print the detailed fidelity report.
    :param mode: The fidelity mode to use: ``"typical"`` (default) or ``"min"``.

    :returns: The estimated fidelity of the circuit as a float.
    :rtype: float
    """
    if verbose:
        run_info_string = "=" * 70 + "\n"
        run_info_string += "Running Quantinuum H2 Fidelity Estimator\n"
        run_info_string += f"Input circuit has {len(circuit.data)} operations\n"
        run_info_string += f"Calculating using '{mode}' infidelity values.\n"
        run_info_string += (
            "Estimating fidelity (transpiling to 'u' and 'cx' gates)...\n"
        )
        run_info_string += "=" * 70
        print(run_info_string)

    # Transpile to 'u' (general 1Q) and 'cx' (general 2Q proxy)
    transpiled_circuit = transpile(
        circuit, basis_gates=["u", "cx"], optimization_level=3
    )

    # Estimate fidelity from the transpiled circuit
    fidelity_result = calculate_circuit_fidelity_quantinuum_h2(
        transpiled_circuit, include_readout=True, mode=mode
    )

    # Print detailed report if requested
    if verbose:
        report_string = get_fidelity_report_string_quantinuum_h2(
            fidelity_result, circuit=transpiled_circuit, original_circuit=circuit
        )
        print(report_string)

    return fidelity_result["fidelity_product"]



import matplotlib.pyplot as plt
import numpy as np
from PIL import Image


def load_image(image_path):
    """
    Load an image from the specified path using PIL

    Parameters:
        image_path (str): Path to the image file

    Returns:
        numpy.ndarray: The loaded image as a numpy array
    """
    pil_image = Image.open(image_path)
    return np.array(pil_image)


def convert_to_grayscale(image):
    """
    Convert an image to grayscale using PIL

    Parameters:
        image (numpy.ndarray): Input RGB image as numpy array

    Returns:
        numpy.ndarray: Grayscale version of the input image
    """
    pil_image = Image.fromarray(np.uint8(image))

    if pil_image.mode != "L":
        pil_image = pil_image.convert("L")

    return np.array(pil_image)


def reshape_image(image, target_height, target_width, keep_aspect_ratio=True):
    """
    Reshape an image to target dimensions using PIL

    Parameters:
        image (numpy.ndarray): Input image as numpy array
        target_height (int): Target height
        target_width (int): Target width
        keep_aspect_ratio (bool): Whether to maintain aspect ratio

    Returns:
        numpy.ndarray: Reshaped image
    """
    pil_image = Image.fromarray(np.uint8(image))

    if keep_aspect_ratio:
        img_width, img_height = pil_image.size
        r = min(target_height / img_height, target_width / img_width)

        new_height = int(img_height * r)
        new_width = int(img_width * r)

        resized = pil_image.resize((new_width, new_height), Image.LANCZOS)

        if pil_image.mode == "L":
            new_image = Image.new("L", (target_width, target_height), 0)
        else:
            new_image = Image.new("RGB", (target_width, target_height), (0, 0, 0))

        y_offset = (target_height - new_height) // 2
        x_offset = (target_width - new_width) // 2

        new_image.paste(resized, (x_offset, y_offset))
        return np.array(new_image)
    resized = pil_image.resize((target_width, target_height), Image.LANCZOS)
    return np.array(resized)


def normalize_image(image):
    """
    Normalize image pixel values to range [0, 1]

    Parameters:
        image (numpy.ndarray): Input image

    Returns:
        numpy.ndarray: Normalized image
    """
    return image.astype(np.float32) / 255.0


def display_image(image, title="Image"):
    """
    Display an image using matplotlib

    Parameters:
        image (numpy.ndarray): Image to display
        title (str): Title for the plot
    """
    if len(image.shape) == 3:
        plt.imshow(image)
    else:
        plt.imshow(image, cmap="gray")
    plt.title(title)
    plt.axis("off")
    plt.show()



# wrapper for random_circuit to have docstring in the desired style
from qiskit import QuantumCircuit
from qiskit.circuit.random import random_circuit as original_random_circuit


def random_circuit(
    num_qubits,
    depth,
    max_operands=4,
    measure=False,
    conditional=False,
    reset=False,
    seed=None,
) -> QuantumCircuit:
    r"""Generate random circuit of arbitrary size and form.

    This function will generate a random circuit by randomly selecting gates
    from the set of standard gates in :mod:`qiskit.extensions`. For example:

    .. plot::
       :include-source:

       from qiskit.circuit.random import random_circuit

       circ = random_circuit(2, 2, measure=True)
       circ.draw(output='mpl')

    :param num_qubits: number of quantum wires
    :param depth: layers of operations (i.e. critical path length)
    :param max_operands: maximum qubit operands of each gate (between 1 and 4)
    :param measure: if True, measure all qubits at the end
    :param conditional: if True, insert middle measurements and conditionals
    :param reset: if True, insert middle resets
    :param seed: sets random seed (optional)
    :raises CircuitError: when invalid options given
    :return: constructed circuit
    """
    return original_random_circuit(
        num_qubits,
        depth,
        max_operands=max_operands,
        measure=measure,
        conditional=conditional,
        reset=reset,
        seed=seed,
    )



from __future__ import annotations

import contextlib
import functools
import logging
import random
import traceback
from pathlib import Path

from bluequbit.bluequbit_logger import BlueQubitLoggerFormatter, PrefixLoggerAdapter

logger_sdk = logging.getLogger("bluequbit-python-sdk")


def error_log_wrapper(func):
    """Wrap a method to emit a best-effort traceback via ``self.logger``."""

    @functools.wraps(func)
    def error_log_wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            *_, (last_frame, _) = traceback.walk_tb(e.__traceback__)
            lines = ["Traceback (most recent call last):\n"]
            lines.extend(traceback.format_stack(last_frame))
            lines.append(type(e).__name__ + ": " + str(e))
            error_info = "".join(lines)
            with contextlib.suppress(Exception):
                args[0].logger.error(error_info)
            raise

    return error_log_wrapped


class ErrorLoggingBase:
    """Apply ``error_log_wrapper`` to subclass callables automatically."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        for name, method in cls.__dict__.items():
            if callable(method):
                setattr(cls, name, error_log_wrapper(method))


class ArtifactLogger:
    """Base class for SDK components with prefixed logging and artifact log files."""

    def __init__(self, artifacts_output_dir: str | Path | None, logger_name: str):
        super().__init__()
        self.artifacts_output_dir: Path | None
        if artifacts_output_dir is not None:
            self.artifacts_output_dir = Path(artifacts_output_dir)
            self.artifacts_output_dir.mkdir(parents=True, exist_ok=True)
            logger = logging.getLogger(
                f"bq_sdk_artifact_{logger_name}_{random.randint(0, 1_000_000_000)}"
            )
            logger.parent = logger_sdk
            logger.propagate = True
            file_handler = logging.FileHandler(
                self.artifacts_output_dir / f"{logger_name}.log", mode="w"
            )
            file_handler.setLevel(logger_sdk.level)
            file_handler.setFormatter(BlueQubitLoggerFormatter(colored=False))
            logger.addHandler(file_handler)
        else:
            logger = logger_sdk
            self.artifacts_output_dir = None
        self.logger = PrefixLoggerAdapter(logger, {"prefix": logger_name})


# Backward compatibility for existing hierarchical-circuit imports.
HierCircuitLogger = ArtifactLogger



from __future__ import annotations

import asyncio
import contextlib
import json
import logging
import os
import time
from typing import TYPE_CHECKING, Any

from . import job_metadata_constants
from .api import circuits, jobs
from .backend_connection import BackendConnection
from .check_version import check_version
from .circuit_serialization import is_a_pennylane_circuit, is_a_qiskit_circuit
from .estimate_result import EstimateResult
from .exceptions import (
    BQBatchJobsLimitExceededError,
    BQCPUJobsLimitExceededError,
    BQJobCouldNotCancelError,
    BQJobInvalidDeviceTypeError,
    BQJobNotCompleteError,
    BQJobsLimitExceededError,
    BQJobsMalformedShotsError,
    BQSDKUsageError,
)
from .job_result import JobResult
from .local_execution import run_circuits_local
from .version import __version__

if TYPE_CHECKING:
    import datetime

# TODO this requires imports of actual quantum libraries for proper type
# checking.
CircuitT = Any

PauliSumT = list[tuple[str, float]] | str

logger = logging.getLogger("bluequbit-python-sdk")


class BQClient:
    """Client for managing jobs on BlueQubit platform.

    :param api_token: API token of the user. If ``None``, the token will be looked
                      in the environment variable BLUEQUBIT_API_TOKEN.
    :param execution_mode: Execution mode. Can be ``"cloud"``, ``"local"`` or ``None``.
                           If ``None``, the environment variable
                           BLUEQUBIT_EXECUTION_MODE will be used if set, otherwise
                           ``"cloud"`` will be used.
    """

    _job_name_prefix_cls: str | None = None

    def __init__(self, api_token: str | None = None, execution_mode: str | None = None):
        super().__init__()
        if os.environ.get("BLUEQUBIT_TESTING") is None:
            with contextlib.suppress(Exception):
                check_version(__version__)

        if execution_mode is None:
            if "BLUEQUBIT_EXECUTION_MODE" in os.environ:
                execution_mode = os.environ["BLUEQUBIT_EXECUTION_MODE"]
                logger.info(
                    "Using execution mode from environment variable BLUEQUBIT_EXECUTION_MODE: %s",
                    execution_mode,
                )
            else:
                execution_mode = "cloud"
        self.execution_mode = execution_mode
        if self.execution_mode not in ["cloud", "local"]:
            raise ValueError(f"Invalid execution mode: {self.execution_mode}")

        self._backend_connection = BackendConnection(api_token)

        self.job_name_prefix: str | None = BQClient._job_name_prefix_cls

    def name(self):
        return "BlueQubit"

    def validate_device(self, device):
        if not isinstance(device, str):
            raise BQJobInvalidDeviceTypeError(device)
        converted_device = device.lower()
        if converted_device not in job_metadata_constants.DEVICE_TYPES:
            raise BQJobInvalidDeviceTypeError(device)
        return converted_device

    @staticmethod
    def validate_batch(batch):
        if not isinstance(batch, list):
            return False
        if len(batch) > job_metadata_constants.MAXIMUM_NUMBER_OF_BATCH_JOBS:
            raise BQBatchJobsLimitExceededError(len(batch))
        return True

    @staticmethod
    def validate_batch_for_run(batch, device):
        if not BQClient.validate_batch(batch):
            return
        if len(batch) > job_metadata_constants.MAXIMUM_NUMBER_OF_JOBS_FOR_RUN:
            raise BQJobsLimitExceededError(len(batch))
        if (
            "cpu" in device
            and len(batch) > job_metadata_constants.QUEUED_CPU_JOBS_LIMIT
        ):
            raise BQCPUJobsLimitExceededError(len(batch))

    @staticmethod
    def validate_circuit_type(circuits, device):
        # Early return for devices that don't require validation
        if not (
            device.startswith("pennylane")
            or device in {"mps.cpu", "mps.gpu", "pauli-path", "pauli-path.gpu"}
        ):
            return

        circuits_list = circuits if isinstance(circuits, list) else [circuits]
        for circuit in circuits_list:
            if device.startswith("pennylane"):
                if not is_a_pennylane_circuit(circuit):
                    raise BQSDKUsageError(
                        f"Only Pennylane circuit is supported for {device}"
                    )
            elif not is_a_qiskit_circuit(circuit):
                raise BQSDKUsageError(f"Only Qiskit circuit is supported for {device}")

    @staticmethod
    def _validate_tags(tags):
        if tags is None:
            return
        if not isinstance(tags, dict):
            raise BQSDKUsageError("'tags' must be a dictionary with keys as strings")
        for k in tags:
            if not isinstance(k, str):
                raise BQSDKUsageError(
                    f"Each key of 'tags' must be a string; received key {k} of type {type(k)}"
                )
        try:
            tags_str = json.dumps(tags)
        except (TypeError, ValueError) as e:
            raise BQSDKUsageError(
                f"'tags' must be JSON serializable; received non-serializable data: {e}"
            ) from e
        tags_size = len(tags_str.encode())
        if tags_size > job_metadata_constants.MAX_JOB_TAGS_SIZE:
            raise BQSDKUsageError(
                f"'tags' must be of size less than 1 MB; received 'tags' of size: {tags_size}"
            )

    @staticmethod
    def _validate_options(options, device):
        if options is None:
            return
        for k in options:
            if k not in [
                "mps_bond_dimension",
                "pauli_path_truncation_threshold",
                "pauli_path_circuit_transpilation_level",
                "pauli_path_weight_cutoff",
            ]:
                raise BQSDKUsageError(f"options not supported: {k}")

        chi = options.get("mps_bond_dimension")
        if chi is not None and (not isinstance(chi, int) or chi < 1):
            raise BQSDKUsageError(
                f"'mps_bond_dimension' must be a positive integer; received {chi}"
            )
        if (
            device == "mps.cpu"
            and chi is not None
            and chi > job_metadata_constants.MAX_MPS_CPU_BOND_DIMENSION
        ):
            raise BQSDKUsageError(
                f"'mps_bond_dimension' must be smaller than {job_metadata_constants.MAX_MPS_CPU_BOND_DIMENSION} for {device}; received {chi}."
            )
        if (
            device == "mps.gpu"
            and chi is not None
            and chi > job_metadata_constants.MAX_MPS_GPU_BOND_DIMENSION
        ):
            raise BQSDKUsageError(
                f"'mps_bond_dimension' must be smaller than {job_metadata_constants.MAX_MPS_GPU_BOND_DIMENSION} for {device}; received {chi}."
            )

        delta = options.get("pauli_path_truncation_threshold")
        if delta is not None and (not isinstance(delta, (float, int)) or delta < 0.0):
            raise BQSDKUsageError(
                f"'pauli_path_truncation_threshold' must be a non-negative float; received {delta}"
            )
        if (
            delta is not None
            and delta < job_metadata_constants.MIN_PAULI_PATH_TRUNCATION_THRESHOLD
        ):
            raise BQSDKUsageError(
                "Truncation threshold too small: 'pauli_path_truncation_threshold' smaller than "
                f"{job_metadata_constants.MIN_PAULI_PATH_TRUNCATION_THRESHOLD}"
                " is currently not supported."
            )
        pps_transpilation_level = options.get("pauli_path_circuit_transpilation_level")
        if (
            pps_transpilation_level is not None
            and pps_transpilation_level
            not in job_metadata_constants.PAULI_PATH_CIRCUIT_TRANSPILATION_LEVELS
        ):
            raise BQSDKUsageError(
                f"'pauli_path_circuit_transpilation_level' must be from {job_metadata_constants.PAULI_PATH_CIRCUIT_TRANSPILATION_LEVELS}; received {pps_transpilation_level}"
            )

    @staticmethod
    def _validate_pps_job(options, circuits, pauli_sum):
        circuits_list = circuits if isinstance(circuits, list) else [circuits]
        # check that if any of the circuits are > 13 qubits, then a truncation threshold is provided
        if options is None or options.get("pauli_path_truncation_threshold") is None:
            for circuit in circuits_list:
                if circuit.num_qubits > 13:
                    raise BQSDKUsageError(
                        "For 'pauli-path' device, circuits with more than 13 qubits require "
                        "'pauli_path_truncation_threshold' to be specified in options."
                        f" Please specify a 'pauli_path_truncation_threshold' larger than {job_metadata_constants.MIN_PAULI_PATH_TRUNCATION_THRESHOLD}."
                    )
        if pauli_sum is None:
            raise BQSDKUsageError(
                "'pauli_sum' must be provided when using the 'pauli-path' device."
            )
        if isinstance(pauli_sum[0], list) and len(pauli_sum) > 1:
            raise BQSDKUsageError(
                "Only a single observable is supported on 'pauli-path' device. "
                f"Received 'pauli_sum' with {len(pauli_sum)} observables."
            )

    def estimate(
        self, circuits: CircuitT | list[CircuitT], device: str = "cpu"
    ) -> EstimateResult | list[EstimateResult]:
        """Estimate job runtime. Not supported in local execution mode.

        :param circuits: quantum circuit or circuits
        :type circuits: Cirq, Qiskit, list
        :param device: device for which to estimate the circuit. Can be one of
                       ``"cpu"`` | ``"gpu"`` | ``"quantum"``
        :return: result or results estimate metadata
        """
        if self.execution_mode == "local":
            raise BQSDKUsageError(
                "BQClient.estimate() is not supported in local execution mode."
            )
        device = self.validate_device(device)
        self.validate_batch(circuits)
        response = jobs.submit_jobs(
            self._backend_connection, circuits, device, estimate_only=True
        )
        if isinstance(circuits, list):
            return [EstimateResult(data) for data in response]
        return EstimateResult(response)

    def run(
        self,
        circuits: CircuitT | list[CircuitT],
        device: str = "cpu",
        *,
        asynchronous: bool = False,
        job_name: str | None = None,
        shots: int | None = None,
        pauli_sum: PauliSumT | list[PauliSumT] | None = None,
        options: dict | None = None,
        tags: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> JobResult | list[JobResult]:
        """Submit a job to run on BlueQubit platform.

        :param circuits: quantum circuit or list of circuits
        :type circuits: Cirq, Qiskit, list
        :param device: device on which to run the circuit. Can be one of
                       ``"cpu"``
                       | ``"gpu"``
                       | ``"quantum"``
                       | ``"mps.cpu"``
                       | ``"mps.gpu"``
                       | ``"pauli-path"``
        :param asynchronous: if set to ``False``, wait for job completion before
                             returning. If set to ``True``, return immediately
        :param job_name: customizable job name
        :param shots: number of shots to run. If device is quantum and shots is None then
                      it is set to 1000. For non quantum devices, if None, full
                      probability distribution will be returned. For mps.cpu and mps.gpu devices it is limited to 1000000, for all other devices it is limited to 100000
        :param pauli_sum: The Pauli sum or a list of Pauli sum which
                          expectation value is the computation result
        :param options: dictionary of options used to configure the run:

                       - ``mps_bond_dimension`` (int): Sets a limit on the number of Schmidt
                         coefficients retained at the end of the svd algorithm.
                         Coefficients beyond this limit will be discarded.
                         Default: None (no limit on the bond dimension).
                         Applicable only to ``"mps.cpu"`` and ``"mps.gpu"`` devices.
                       - ``pauli_path_truncation_threshold`` (float): Sets a lower bound on the absolute value of
                         coefficients of the Pauli operators retained after each step of the Pauli-path propagation algorithm.
                         Pauli operators with coefficients less than this threshold will be discarded.
                         Choosing a smaller threshold keeps more Pauli terms in the simulation and provides
                         a better estimate of the expectation value, but at the cost of increased runtime.
                         Smallest value currently supported is 1e-5.
                         For circuits with more than 13 qubits a ``pauli_path_truncation_threshold`` must be specified.
                         For exact simulation of circuits with <= 13 qubits, set ``pauli_path_truncation_threshold`` to ``None``.
                         Applicable only to ``"pauli-path"`` device.
                       - ``pauli_path_circuit_transpilation_level`` (int): Controls the extend of transpilation performed
                         while converting the given circuit into one consisting of only rotation gates.
                         Must be one of 0, 1, 2 or 3, with 0 corresponding to minimal tranpilation, whereas level 3 involves
                         extensive, potentially time consuming, transpilation passes with different bases to minimize gate counts.
                         In general, different transpilation levels may result in slightly different expectation values.
                         Default: 2
                         Applicable only to ``"pauli-path"`` device

        :param tags: an optional dictionary containing key-value pairs to help tag the job.

        :param timeout: maximum wait time in seconds when ``asynchronous`` is ``False``.
                        If the job does not complete within the specified time,
                        an exception will be raised. If None, the wait will continue indefinitely.

        :return: job or jobs metadata
        """
        device = self.validate_device(device)
        self.validate_batch_for_run(circuits, device)
        self.validate_circuit_type(circuits, device)
        self._validate_options(options, device)
        self._validate_tags(tags)
        if timeout is not None:
            if self.execution_mode == "local":
                raise BQSDKUsageError(
                    "BQClient.run(timeout=...) is not supported in local execution mode."
                )
            if asynchronous is True:
                raise BQSDKUsageError(
                    "BQClient.run(asynchronous=True) does not support timeout parameter."
                )
        if device == "pauli-path":
            self._validate_pps_job(options, circuits, pauli_sum)

        if shots is not None and (
            not isinstance(shots, int)
            or shots > job_metadata_constants.MAXIMUM_NUMBER_OF_SHOTS[device]
        ):
            raise BQJobsMalformedShotsError(
                shots, device, job_metadata_constants.MAXIMUM_NUMBER_OF_SHOTS[device]
            )
        job_name = (
            job_name
            if self.job_name_prefix is None
            else self.job_name_prefix + str(job_name)
        )
        if self.execution_mode == "local" and device != "quantum":
            if asynchronous is True:
                raise BQSDKUsageError(
                    "BQClient.run(asynchronous=True) is not supported in local execution mode."
                )
            return run_circuits_local(
                circuits,
                device,
                job_name=job_name,
                shots=shots,
                pauli_sum=pauli_sum,
                options=options,
                tags=tags,
            )
        if job_name is not None:
            if not isinstance(circuits, list):
                if len(job_name) > job_metadata_constants.MAX_JOB_NAME_LENGTH:
                    logger.warning(
                        "Job name is too long, it will be truncated to %s characters.",
                        job_metadata_constants.MAX_JOB_NAME_LENGTH,
                    )
            else:  # batch
                zero_fill_count = len(str(len(circuits) - 1))
                if (
                    len(job_name)
                    > job_metadata_constants.MAX_JOB_NAME_LENGTH - zero_fill_count
                ):
                    logger.warning(
                        "Batch job name is too long, it will be truncated to %s characters.",
                        job_metadata_constants.MAX_JOB_NAME_LENGTH,
                    )
        response = jobs.submit_jobs(
            self._backend_connection,
            circuits,
            device,
            job_name,
            shots=shots,
            asynchronous=asynchronous,
            pauli_sum=pauli_sum,
            options=options,
            tags=tags,
        )
        if isinstance(circuits, list):
            logger.info(
                "Submitted %s jobs. Batch ID %s", len(response), response[0]["batch_id"]
            )

            def add_circuit_to_all(job_results, circuits):
                for job_result, circuit in zip(job_results, circuits, strict=True):
                    job_result.circuit = circuit

            job_results = [
                JobResult(data, self._backend_connection) for data in response
            ]
            if not asynchronous:
                if self._check_all_in_terminal_states(job_results):
                    add_circuit_to_all(job_results, circuits)
                    return job_results

                waited_job_results = self.wait(job_results, timeout=timeout)
                add_circuit_to_all(waited_job_results, circuits)
                return waited_job_results
                # if job_results[0].batch_id is not None:
                #     return self.wait(batch_id=job_results[0].batch_id)
                # else:
            add_circuit_to_all(job_results, circuits)
            return job_results
        submitted_job = JobResult(response, self._backend_connection)
        if (
            submitted_job.run_status
            in job_metadata_constants.JOB_NO_RESULT_TERMINAL_STATES
        ):
            raise BQJobNotCompleteError(
                submitted_job.job_id,
                submitted_job.run_status,
                submitted_job.error_message,
            )
        logger.info("Submitted: %s", submitted_job)
        if (
            not asynchronous
            and submitted_job.run_status
            not in job_metadata_constants.JOB_TERMINAL_STATES
        ):
            jr = self.wait(submitted_job.job_id, timeout=timeout)
            jr.circuit = circuits  # type: ignore[union-attr]
            return jr
        submitted_job.circuit = circuits
        return submitted_job

    @staticmethod
    def _check_all_in_terminal_states(job_results):
        if not isinstance(job_results, list):
            if (
                job_results.run_status
                in job_metadata_constants.JOB_NO_RESULT_TERMINAL_STATES
            ):
                raise BQJobNotCompleteError(
                    job_results.job_id,
                    job_results.run_status,
                    job_results.error_message,
                )
            return job_results.run_status in job_metadata_constants.JOB_TERMINAL_STATES
        for job_result in job_results:
            if (
                job_result.run_status
                in job_metadata_constants.JOB_NO_RESULT_TERMINAL_STATES
            ):
                raise BQJobNotCompleteError(
                    job_result.job_id,
                    job_result.run_status,
                    job_result.error_message,
                )
        return all(
            job_result.run_status in job_metadata_constants.JOB_TERMINAL_STATES
            for job_result in job_results
        )

    def wait(
        self,
        job_ids: str | list[str] | JobResult | list[JobResult],
        timeout: float | None = None,
    ) -> JobResult | list[JobResult]:
        """Wait for job completion. Not supported in local execution mode.

        :param job_ids: job IDs that can be found as property of :class:`JobResult` metadata
                        of :func:`~run` method, or `JobResult` instances from which job IDs
                        will be extracted
        :param timeout: maximum wait time in seconds
                             If the job does not complete within the specified time,
                             an exception will be raised. If None, the wait will continue indefinitely. The raised error will have an error.job_ids attribute that can be used to identify the jobs that didn't complete within the specified time and be canceled.
        :return: job metadata
        """
        if self.execution_mode == "local":
            raise BQSDKUsageError(
                "BQClient.wait() is not supported in local execution mode."
            )
        self.validate_batch(job_ids)
        job_results_previous = None
        start_time = time.time()
        while True:
            job_results = self._get(job_ids, need_qc_unprocessed=False)
            if self._check_all_in_terminal_states(job_results):
                return job_results
            if not isinstance(job_results, list):
                job_results = [job_results]
            if job_results_previous is not None:
                for job_result, job_result_previous in zip(
                    job_results, job_results_previous, strict=True
                ):
                    assert job_result.job_id == job_result_previous.job_id
                    if job_result.run_status != job_result_previous.run_status:
                        logger.info(
                            "Status changed: Job ID: %s. From %s to %s.",
                            job_result.job_id,
                            job_result_previous.run_status,
                            job_result.run_status,
                        )
            job_results_previous = job_results
            time.sleep(1.0)
            if timeout is not None and time.time() - start_time > timeout:
                error_job_ids = (
                    [jr.job_id for jr in job_results]
                    if isinstance(job_ids, list)
                    else job_results[0].job_id
                )
                err = TimeoutError(
                    f"Jobs {error_job_ids} didn't complete within the specified time. Maximum wait time: {timeout} seconds.",
                    timeout,
                )
                err.job_ids = error_job_ids  # type: ignore[attr-defined]
                raise err

    def get(
        self,
        job_ids: str | list[str] | JobResult | list[JobResult],
    ) -> JobResult | list[JobResult]:
        """Get current metadata of jobs. Not supported in local execution mode.

        :param job_ids: job IDs that can be found as property of :class:`JobResult` metadata
                        of :func:`~run` method
        :return: jobs metadata
        """
        if self.execution_mode == "local":
            raise BQSDKUsageError(
                "BQClient.get() is not supported in local execution mode."
            )
        return self._get(job_ids)

    def _get(
        self,
        job_ids: str | list[str] | JobResult | list[JobResult],
        need_qc_unprocessed=True,
    ) -> JobResult | list[JobResult]:
        self.validate_batch(job_ids)
        job_ids_list = job_ids if isinstance(job_ids, list) else [job_ids]
        if isinstance(job_ids_list[0], JobResult):
            job_ids_list = [jr.job_id for jr in job_ids_list]  # type: ignore[union-attr]
        job_results = jobs.search_jobs(
            self._backend_connection,
            job_ids=job_ids_list,
            need_qc_unprocessed=need_qc_unprocessed,
        )
        job_results = [
            JobResult(r, self._backend_connection) for r in job_results["data"]
        ]
        if isinstance(job_ids, list):
            return job_results
        return job_results[0]

    def cancel(
        self, job_ids: str | list[str] | JobResult | list[JobResult]
    ) -> JobResult | list[JobResult]:
        """Submit jobs cancel request. Not supported in local execution mode.

        :param job_ids: job IDs that can be found as property of :class:`JobResult` metadata
                        of :func:`run` method
        :return: job or jobs metadata
        """
        if self.execution_mode == "local":
            raise BQSDKUsageError(
                "BQClient.cancel() is not supported in local execution mode."
            )
        self.validate_batch(job_ids)
        if isinstance(job_ids, JobResult):
            job_ids = job_ids.job_id
        elif isinstance(job_ids, list) and isinstance(job_ids[0], JobResult):
            job_ids = [jr.job_id for jr in job_ids]  # type: ignore[union-attr]
        logger.info("Canceling jobs: %s", job_ids)
        responses = jobs.cancel_jobs(self._backend_connection, job_ids)
        if isinstance(job_ids, list):
            for response in responses:
                if response["ret"] == "FAILED":
                    logger.warning(response["error_message"])
        try:
            self.wait(job_ids, timeout=180.0)
        except BQJobNotCompleteError as e:
            if not e.run_status == "CANCELED":
                raise BQJobCouldNotCancelError(
                    e.job_id, e.run_status, e.error_message
                ) from None
        return self.get(job_ids)

    def search(
        self,
        run_status: str | None = None,
        created_later_than: str | datetime.datetime | None = None,
        batch_id: str | None = None,
    ) -> list[JobResult]:
        """Search jobs. Not supported in local execution mode.

        :param run_status: if not ``None``, run status of jobs to filter.
                           Can be one of ``"FAILED_VALIDATION"`` | ``"PENDING"`` |
                           ``"QUEUED"`` | ``"RUNNING"`` | ``"TERMINATED"`` | ``"CANCELED"`` |
                           ``"NOT_ENOUGH_FUNDS"`` | ``"COMPLETED"``

        :param created_later_than: if not ``None``, filter by latest job creation datetime.
                                   Please add timezone for clarity, otherwise UTC
                                   will be assumed

        :param batch_id: if not ``None``, filter by batch ID

        :return: metadata of jobs
        """
        if self.execution_mode == "local":
            raise BQSDKUsageError(
                "BQClient.search() is not supported in local execution mode."
            )
        job_results = jobs.search_jobs(
            self._backend_connection, run_status, created_later_than, batch_id=batch_id
        )
        return [JobResult(r, self._backend_connection) for r in job_results["data"]]

    def get_peaked_circuit(self, difficulty: int, circuit_id: int | None = None):
        """Get a peaked circuit from the backend.

        :param difficulty: difficulty of the circuit to get
        :param circuit_id: ID of the circuit to get. If ``None``, a random
                           circuit of the given difficulty will be returned.
        :return: PeakedCircuit
        :raises BQSDKUsageError: if circuit_id is not a non-negative integer
        """
        from .peaked_circuit import PeakedCircuit

        if circuit_id is not None and (
            not isinstance(circuit_id, int) or circuit_id < 0
        ):
            raise BQSDKUsageError("circuit_id must be a non-negative integer")
        if difficulty is None or not isinstance(difficulty, int) or difficulty <= 0:
            raise BQSDKUsageError("difficulty must be a positive integer")
        data = circuits.get_peaked_circuit(
            self._backend_connection, difficulty, circuit_id
        )
        return PeakedCircuit(data)

    def _set_api_min_request_interval(self, min_request_interval: float) -> None:
        """Set the minimum request interval for the API. If consecutive requests are made
        within this interval, the SDK will automatically wait until the interval has passed
        before making the next request.

        :param min_request_interval: minimum request interval in seconds
        :return: None
        """
        if min_request_interval < 0.0:
            raise ValueError("Minimum request interval must be non-negative")
        self._backend_connection.min_request_interval = min_request_interval

    def _get_api_min_request_interval(self) -> float:
        """Get the minimum request interval for the API. See :func:`set_api_min_request_interval` for more details.

        :return: minimum request interval in seconds
        """
        return self._backend_connection.min_request_interval

    async def run_native_async(
        self, *args, polling_interval: float = 1.0, **kwargs
    ) -> JobResult | list[JobResult]:
        """Experimental: Submit a job to run on BlueQubit platform with asyncio.
        Not supported in local execution mode.

        :param args: arguments for the :func:`run` method
        :param polling_interval: interval in seconds to poll for job completion
        :type polling_interval: float
        :param kwargs: keyword arguments for the :func:`run` method

        :return: job or jobs metadata
        :rtype: JobResult or list[JobResult]
        """
        if self.execution_mode == "local":
            raise BQSDKUsageError("asyncio is not supported in local execution mode.")
        if kwargs.get("asynchronous") is False:
            raise BQSDKUsageError("asyncio is not supported in not asynchronous mode.")
        job_results = self.run(*args, asynchronous=True, **kwargs)
        try:
            while True:
                job_results = self._get(job_results, need_qc_unprocessed=False)

                if self._check_all_in_terminal_states(job_results):
                    return job_results
                await asyncio.sleep(polling_interval)
        except asyncio.CancelledError:
            job_ids = (
                [jr.job_id for jr in job_results]
                if isinstance(job_results, list)
                else [job_results.job_id]
            )
            logger.info("Cancelling jobs: %s", job_ids)
            jobs.cancel_jobs(self._backend_connection, job_ids)
            raise



from __future__ import annotations

import numpy as np


def format_floats_for_display(
    obj: dict | list | float | np.floating, decimal_places: int = 2
) -> dict | list | str:
    """
    Recursively format floats in a data structure to scientific notation with specified decimal places

    Parameters:
        obj: The object to format (dict, list, float, etc.)
        decimal_places (int): Number of decimal places in scientific notation (default: 2)

    Returns:
        Formatted object with floats converted to scientific notation strings
    """
    if isinstance(obj, dict):
        return {k: format_floats_for_display(v, decimal_places) for k, v in obj.items()}
    if isinstance(obj, list):
        return [format_floats_for_display(item, decimal_places) for item in obj]
    if isinstance(obj, float):
        return f"{obj:.{decimal_places}e}"
    if isinstance(obj, np.floating):
        return f"{float(obj):.{decimal_places}e}"
    return obj



from __future__ import annotations

import datetime
import json
import logging
import zlib

import dateutil.parser

from bluequbit.circuit_serialization import encode_circuit_with_fallback
from bluequbit.exceptions import (
    BQError,
    BQJobExceedsSerializedCircuitSizeLimitError,
)
from bluequbit.job_metadata_constants import MAXIMUM_SERIALIZED_CIRCUIT_SIZE
from bluequbit.tools import format_pauli_sum

logger = logging.getLogger("bluequbit-python-sdk")

_SINGLE_REQUEST_LIMIT = 100


def search_jobs(
    _connection,
    run_status=None,
    created_later_than=None,
    job_ids=None,
    batch_id=None,
    *,
    need_qc_unprocessed=False,
):
    if created_later_than is None:
        parsed_created_later_than = None
    elif isinstance(created_later_than, str):
        parsed_created_later_than_datetime = dateutil.parser.parse(created_later_than)
        if parsed_created_later_than_datetime.tzinfo is None:
            logger.warning(
                "created_later_than is a str object without timezone info, assuming UTC"
                " timezone"
            )
            parsed_created_later_than_datetime = (
                parsed_created_later_than_datetime.replace(tzinfo=datetime.timezone.utc)
            )
        parsed_created_later_than = parsed_created_later_than_datetime.isoformat()
    elif isinstance(created_later_than, datetime.datetime):
        if created_later_than.tzinfo is None:
            raise BQError(
                "created_later_than is a datetime object without timezone info"
            )
        parsed_created_later_than = created_later_than.isoformat()
    else:
        raise BQError(
            "created_later_than should be None, str, or datetime.datetime object"
        )

    params = {
        "limit": _SINGLE_REQUEST_LIMIT,
        "run_status": run_status,
        "created_later_than": parsed_created_later_than,
        "job_ids": job_ids,
        "batch_id": batch_id,
        "need_qc_unprocessed": str(need_qc_unprocessed),
        # "need_top_128": True,
    }
    result_dict: dict = {"data": []}
    while True:
        response = _connection.send_request(req_type="GET", path="/jobs", params=params)
        response = response.json()

        results = response["data"]
        result_dict["data"] += results
        result_dict["total_count"] = response["total_count"]

        if len(results) < _SINGLE_REQUEST_LIMIT:
            break

        params["offset"] = len(result_dict["data"])
    return result_dict


def submit_jobs(
    _connection,
    circuits,
    device,
    job_name=None,
    shots=None,
    pauli_sum=None,
    options=None,
    tags=None,
    *,
    estimate_only=False,
    asynchronous=False,
):
    encoded_circuits: dict | list[dict]
    if isinstance(circuits, list):
        encoded_circuits = [
            encode_circuit_with_fallback(circuit) for circuit in circuits
        ]
        if pauli_sum is not None:
            for ec in encoded_circuits:
                ec["pauli_sum"] = format_pauli_sum(pauli_sum)
        if options is not None:
            for ec in encoded_circuits:
                ec["options"] = options
    else:
        encoded_circuits = encode_circuit_with_fallback(circuits)
        assert isinstance(encoded_circuits, dict)
        if pauli_sum is not None:
            encoded_circuits["pauli_sum"] = format_pauli_sum(pauli_sum)
        if options is not None:
            encoded_circuits["options"] = options

    params = {
        "estimate_only": estimate_only,
        "circuit": encoded_circuits,
        "job_name": job_name,
        "device": device,
        "shots": shots,
        "asynchronous": asynchronous,
        "tags": tags,
    }
    data_str = json.dumps(params)
    data = data_str.encode()
    data_size = len(data)
    if data_size > MAXIMUM_SERIALIZED_CIRCUIT_SIZE:
        raise BQJobExceedsSerializedCircuitSizeLimitError(data_size)

    if data_size > 2000:
        data = zlib.compress(data)
        headers = {"content-encoding": "gzip"}
    else:
        headers = {"content-type": "application/json"}
    response = _connection.send_request(
        req_type="POST", path="/jobs", data=data, headers=headers
    )
    return response.json()["data"]


def cancel_jobs(_connection, job_ids):
    response = _connection.send_request(
        req_type="PATCH", path="/jobs", json_req={"job_ids": job_ids, "cancel": True}
    )
    return response.json()["data"]



"""
Script to fetch and save IBM FEZ Quantum backend calibration data to ``.json.gz``.

This script connects to IBM Quantum using your token and saves all the
necessary backend calibration data (gate fidelities, measurement errors, etc.)
to a ``.json.gz`` file. This data can then be used for offline fidelity estimation
without requiring an IBM token.
"""

import gzip
import json
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from qiskit_ibm_runtime import QiskitRuntimeService


def extract_backend_data(backend) -> dict[str, Any]:
    """Extract all necessary calibration and transpilation data from a backend.

    :param backend: IBM backend object.

    :returns: Dictionary containing all backend calibration and transpilation data.
    :rtype: dict
    """
    creation_datetime = datetime.now(timezone.utc).isoformat()

    target = backend.target

    # Extract gate error data
    gate_errors: dict[str, dict[str, dict[str, Any]]] = {}
    for gate_name in target.operation_names:
        if gate_name in target:
            gate_errors[gate_name] = {}
            for qubits in target[gate_name]:
                if qubits is not None:
                    props = target[gate_name][qubits]

                    # Skip if props is None (can happen for some gates)
                    if props is None:
                        continue

                    # Convert tuple to string for JSON serialization
                    qubit_key = (
                        str(qubits) if isinstance(qubits, tuple) else str((qubits,))
                    )

                    # Handle duration - could be float or object with .magnitude attribute
                    duration = None
                    if props.duration is not None:
                        if hasattr(props.duration, "magnitude"):
                            duration = props.duration.magnitude
                        else:
                            duration = props.duration

                    gate_errors[gate_name][qubit_key] = {
                        "error": props.error,
                        "duration": duration,
                    }

    # Extract qubit properties (T1, T2, frequency, etc.)
    qubit_properties = {}
    for qubit in range(backend.num_qubits):
        if target.qubit_properties and target.qubit_properties[qubit]:
            props = target.qubit_properties[qubit]
            qubit_properties[str(qubit)] = {
                "t1": props.t1,
                "t2": props.t2,
                "frequency": props.frequency if hasattr(props, "frequency") else None,
            }

    # Extract coupling map (critical for transpilation)
    coupling_map = []
    if hasattr(backend, "coupling_map") and backend.coupling_map:
        # Convert EdgeList to regular Python list for JSON serialization
        coupling_map = list(backend.coupling_map.get_edges())

    # Extract basis gates
    basis_gates = list(target.operation_names)

    # Extract backend configuration details
    dt = None
    if hasattr(backend, "dt"):
        dt = backend.dt

    max_shots = None
    if hasattr(backend, "max_shots"):
        max_shots = backend.max_shots

    # Create comprehensive backend data structure
    backend_data = {
        "creation_datetime": creation_datetime,
        "backend_name": backend.name,
        "num_qubits": backend.num_qubits,
        "gate_errors": gate_errors,
        "qubit_properties": qubit_properties,
        "operation_names": list(target.operation_names),
        # Transpilation data
        "coupling_map": coupling_map,
        "basis_gates": basis_gates,
        "dt": dt,
        "max_shots": max_shots,
    }

    return backend_data


def save_backend_data(backend, output_file: str | None = None):
    """Save backend calibration data to a JSON file.

    :param backend: IBM backend object.
    :param output_file: Path to the output JSON file (default: ``{backend_name}_calibration.json``).
    """
    if output_file is None:
        output_file = f"{backend.name}_calibration.json.gz"

    print(f"Extracting calibration data from {backend.name}...")
    backend_data = extract_backend_data(backend)

    # Save to JSON
    print(f"Saving to {output_file}...")
    with gzip.open(output_file, "wt") as f:
        json.dump(backend_data, f, indent=2)

    print(f"✓ Successfully saved backend data to {output_file}")
    print(f"  Backend: {backend_data['backend_name']}")
    print(f"  Qubits: {backend_data['num_qubits']}")
    print(f"  Operations: {len(backend_data['operation_names'])}")
    print(f"  Coupling map edges: {len(backend_data['coupling_map'])}")
    print(f"  File size: {Path(output_file).stat().st_size / 1024:.2f} KB")
    print(f"  Creation datetime: {backend_data['creation_datetime']}")
    print("\n✓ Saved data includes:")
    print("  - Gate fidelities and errors")
    print("  - Qubit properties (T1, T2, frequency)")
    print("  - Coupling map (for transpilation)")
    print("  - Basis gates and backend configuration")
    print("  - Creation datetime")


def main():
    """Main function to connect to IBM Quantum and save backend data.

    This function connects to IBM Quantum using the provided token and instance,
    retrieves the backend data, and saves it to a ``.json.gz`` file.
    """
    # Configuration - Replace with your credentials
    ibm_token = ""  # Your IBM Quantum token
    ibm_instance = "crn:v1:bluemix:public:quantum-computing:us-east:a/373146ee8c834aa5a059a09701d8d31b:fd7f08b1-6d70-442e-bf8b-6df950815350::"
    backend_name = "ibm_fez"

    if not ibm_token:
        print("ERROR: Please set your ibm_token in the script")
        return

    print("=" * 70)
    print("IBM FEZ Quantum Backend Data Saver")
    print("=" * 70)
    print()

    # Connect to IBM Quantum
    print("Connecting to IBM Quantum...")
    service = QiskitRuntimeService(
        channel="ibm_cloud", token=ibm_token, instance=ibm_instance
    )

    # Get backend
    print(f"Getting backend: {backend_name}...")
    backend = service.backend(backend_name)

    # Save data
    save_backend_data(backend)

    print()
    print("=" * 70)
    print("Done! You can now use this file for offline fidelity estimation.")
    print("=" * 70)


if __name__ == "__main__":
    main()



# Qiskit notice begins here
# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# Qiskit notice ends here

# This file has been modified based on:
# https://github.com/Qiskit/qiskit/blob/main/qiskit/circuit/library/arithmetic/adders/draper_qft_adder.py
# Changes made:
# - The class has been renamed from DraperQFTAdder to DraperAdder
# - Modified to support addition of registers with different sizes, instead of equally sized registers
# - Changed the base class from Adder to QuantumCircuit, because Adder is for equally sized registers
# - Added arguments for the sizes of each register. The second one is optional and equal to the first by default
# - Added two arguments  that allow to remove QFT and IQFT parts of the circuit
# - Changed the name of the circuit to this form: "DraperAdder_m_n_kind_int(include_qft)_int(include_iqft))"
# - Changed the docstring to adjust the style and correspond to the modifications


"""Compute the sum of two qubit registers using QFT."""

from __future__ import annotations

import numpy as np
from qiskit.circuit import QuantumRegister
from qiskit.circuit.library.basis_change import QFT
from qiskit.circuit.quantumcircuit import QuantumCircuit


class DraperAdder(QuantumCircuit):
    r"""A circuit that uses QFT to perform in-place addition on two qubit registers.

    For registers with :math:`m` and :math:`n` qubits, where :math:`m \leq n`, the QFT adder can perform addition modulo
    :math:`2^n` (with ``kind="fixed"``) or ordinary addition by adding a carry qubits (with
    ``kind="half"``). It also provides options to omit ``QFT`` or ``IQFT`` parts of DraperAdder through
    ``include_qft`` and ``include_iqft`` arguments.

    The name of the circuit is in the form "DraperAdder_m_n_kind_int(include_qft)_int(include_iqft)",
    e.g. DraperAdder_3_5_half_1_0

    :param m: The number of qubits in the first input register for
            state :math:`|a\rangle`. The second input
            registers must have at least as many qubits as the first one (m<=n).
    :param n: The number of qubits in the second input register for
            state :math:`|b\rangle`. The second input
            registers must have at least as many qubits as the first one (m<=n).
            If ``n`` is not provided it will be equal to ``m`` by default.
    :param kind: The kind of adder, can be ``'half'`` for a half adder or
            ``'fixed'`` for a fixed-sized adder. A half adder contains a carry-out to represent
            the most-significant bit, but the fixed-sized adder doesn't and hence performs
            addition modulo ``2 ** num_state_qubits``.
    :param name: The name of the circuit object \(default is "DraperAdder"\). The other parameters will be added to
        this name as described above
    :param include_qft: If ``False``, the ``QFT`` part is not added to the circuit
    :param include_iqft: If ``False``, the ``IQFT`` part is not added to the circuit
    :raises ValueError: If ``m`` is lower than 1 or if ``m > n``


    As an example, a non-fixed_point QFT adder circuit that performs addition on two 2-qubit sized
    registers is as follows:

    .. parsed-literal::

         a_0:   ─────────■──────■────────────────────────■────────────────
                         │      │                        │
         a_1:   ─────────┼──────┼────────■──────■────────┼────────────────
                ┌──────┐ │P(π)  │        │      │        │       ┌───────┐
         b_0:   ┤0     ├─■──────┼────────┼──────┼────────┼───────┤0      ├
                │      │        │P(π/2)  │P(π)  │        │       │       │
         b_1:   ┤1 qft ├────────■────────■──────┼────────┼───────┤1 iqft ├
                │      │                        │P(π/2)  │P(π/4) │       │
        cout_0: ┤2     ├────────────────────────■────────■───────┤2      ├
                └──────┘                                         └───────┘

    **References:**

    [1] T. G. Draper, Addition on a Quantum Computer, 2000.
    `arXiv:quant-ph/0008033 <https://arxiv.org/pdf/quant-ph/0008033.pdf>`_

    [2] Ruiz-Perez et al., Quantum arithmetic with the Quantum Fourier Transform, 2017.
    `arXiv:1411.5949 <https://arxiv.org/pdf/1411.5949.pdf>`_

    [3] Vedral et al., Quantum Networks for Elementary Arithmetic Operations, 1995.
    `arXiv:quant-ph/9511018 <https://arxiv.org/pdf/quant-ph/9511018.pdf>`_

    """

    def __init__(
        self,
        m: int,
        n: int | None = None,
        kind: str = "fixed",
        name: str = "DraperAdder",
        include_qft: bool = True,
        include_iqft: bool = True,
    ) -> None:
        if n is None:
            n = m

        if kind == "full":
            raise ValueError(
                "The DraperAdder only supports 'half' and 'fixed' as ``kind``."
            )

        if m < 1:
            raise ValueError(
                "The number of qubits in the first register must be at least 1."
            )

        if m > n:
            raise ValueError(
                "The first register must not have more qubits than the second (m <= n)"
            )

        name += f"_{m}_{n}_{kind}_{int(include_qft)}_{int(include_iqft)}"
        super().__init__(name=name)

        qr_a = QuantumRegister(m, name="a")
        qr_b = QuantumRegister(n, name="b")
        qr_list = [qr_a, qr_b]

        if kind == "half":
            qr_z = QuantumRegister(1, name="cout")
            qr_list.append(qr_z)

        # add registers
        self.add_register(*qr_list)

        # define register containing the sum and number of qubits for QFT circuit
        qr_sum = qr_b[:] if kind == "fixed" else qr_b[:] + qr_z[:]
        num_qubits_qft = n if kind == "fixed" else n + 1

        circuit = QuantumCircuit(*self.qregs, name=name)

        # build QFT adder circuit
        if include_qft:
            circuit.append(QFT(num_qubits_qft, do_swaps=False).to_gate(), qr_sum[:])

        for j in range(m):
            for k in range(n - j):
                lam = np.pi / (2**k)
                circuit.cp(lam, qr_a[j], qr_b[j + k])

        if kind == "half":
            for j in range(m):
                lam = np.pi / (2 ** (j + 1 + n - m))
                circuit.cp(lam, qr_a[m - j - 1], qr_z[0])

        if include_iqft:
            circuit.append(
                QFT(num_qubits_qft, do_swaps=False).inverse().to_gate(), qr_sum[:]
            )

        self.append(circuit.to_gate(), self.qubits)



from .fidelity_estimation import estimate_circuit_fidelity
from .ibm_fez_fidelity_estimation import estimate_circuit_fidelity_ibm_fez
from .quantinuum_h2_fidelity_estimation import estimate_circuit_fidelity_quantinuum_h2

__all__ = [
    "estimate_circuit_fidelity",
    "estimate_circuit_fidelity_ibm_fez",
    "estimate_circuit_fidelity_quantinuum_h2",
]



"""PeakedCircuit result object returned by BQClient.get_peaked_circuit()."""

from __future__ import annotations

from qiskit import QuantumCircuit


class PeakedCircuit:
    """Represents a peaked circuit fetched from the BlueQubit backend.

    Attributes:
        circuit: The Qiskit QuantumCircuit reconstructed from the QASM string.
        key: The peak bitstring associated with the circuit.
        metadata: A string describing the version or provenance of the circuit.
        qasm_str: The raw QASM string of the circuit.
    """

    def __init__(self, data: dict[str, str]) -> None:
        """Initialize a PeakedCircuit from the backend response dict.

        :param data: Dictionary with keys ``qasm_str``, ``key``, and ``metadata``.
        """
        #: str: raw QASM string
        self.qasm_str: str = data["qasm_str"]

        #: QuantumCircuit: the reconstructed Qiskit circuit
        self.circuit: QuantumCircuit = QuantumCircuit.from_qasm_str(self.qasm_str)

        #: str: peak bitstring
        self.key: str = data["key"]

        #: str: version/provenance metadata
        self.metadata: str = data["metadata"]

    def __repr__(self) -> str:
        return (
            f"PeakedCircuit(key={self.key!r}, metadata={self.metadata!r}, "
            f"num_qubits={self.circuit.num_qubits})"
        )



import logging
import math

import numpy as np

logger = logging.getLogger("bluequbit-python-sdk")


def find_largest_power_of_2(n):
    """
    Find the largest power of 2 that is less than or equal to n

    Parameters:
        n (int): Input number

    Returns:
        int: Largest power of 2 <= n
    """
    if n <= 0:
        return 1

    power = int(math.log2(n))
    return 2**power


def find_best_grid_arrangement(num_blocks, height, width):
    """
    Find the best grid arrangement (blocks_y, blocks_x) for a given number of blocks
    that maximizes the resulting block size (minimum of block height and block width)

    Parameters:
        num_blocks (int): Target number of blocks
        height (int): Image height
        width (int): Image width

    Returns:
        tuple: (blocks_y, blocks_x) that gives the largest possible blocks
    """
    best_arrangement = (1, num_blocks)
    best_block_size = 0

    for blocks_y in range(1, num_blocks + 1):
        if num_blocks % blocks_y != 0:
            continue
        blocks_x = num_blocks // blocks_y

        block_height = height // blocks_y
        block_width = width // blocks_x

        max_square_size = min(block_height, block_width)

        if max_square_size > best_block_size:
            best_block_size = max_square_size
            best_arrangement = (blocks_y, blocks_x)

    if height % best_arrangement[0] != 0 or width % best_arrangement[1] != 0:
        raise ValueError(
            f"Couldn't split image ({height}x{width}) into {num_blocks} blocks."
        )

    return best_arrangement


def apply_epsilon_clipping(block, epsilon=1e-12):
    """
    Apply epsilon clipping to a block to handle zero values

    Parameters:
        block (numpy.ndarray): Input block
        epsilon (float): Small value to replace zeros with

    Returns:
        tuple: (clipped_block, statistics)
    """
    zero_mask = block == 0
    zeros_count = np.sum(zero_mask)

    clipped_block = block.copy()
    clipped_block[zero_mask] = epsilon

    stats = {
        "zeros_clipped": int(zeros_count),
        "total_pixels": int(block.size),
        "zero_percentage": float(zeros_count / block.size * 100),
        "min_value_before": float(np.min(block)),
        "min_value_after": float(np.min(clipped_block)),
        "max_value": float(np.max(block)),
    }

    return clipped_block, stats


def split_image_for_quantum(image, num_blocks, *, normalize_blocks=True, epsilon=1e-12):
    """
    Split an image into square blocks with power-of-2 dimensions for quantum processing
    Each block is normalized to sum to 1 individually, with epsilon clipping for zero values

    Parameters:
        image (numpy.ndarray): Input image
        num_blocks (int): Number of blocks to create
        normalize_blocks (bool): Whether to normalize each block to sum to 1
        epsilon (float): Small value to clip zero pixels to avoid numerical issues

    Returns:
        dict: Dictionary containing:
            - 'blocks': List of square image blocks (each normalized to sum to 1)
            - 'normalization_factors': List of normalization factors for each block
            - 'epsilon_info': Information about epsilon clipping applied
            - 'block_size': Size of each square block (d)
            - 'qubits_needed': Number of qubits needed per block (2 * log2(d))
            - 'blocks_y': Number of blocks in y direction
            - 'blocks_x': Number of blocks in x direction
            - 'original_shape': Original image shape
            - 'cropped_shape': Shape after cropping to fit blocks
            - 'metadata': Additional info for reconstruction
    """
    # Get image dimensions
    if len(image.shape) == 3:
        height, width, channels = image.shape
    else:
        height, width = image.shape
        channels = 1

    # Find the best grid arrangement for the given number of blocks
    blocks_y, blocks_x = find_best_grid_arrangement(num_blocks, height, width)

    # Calculate the maximum possible block dimensions
    max_block_height = height // blocks_y
    max_block_width = width // blocks_x

    # Find the largest power of 2 that fits in both dimensions
    max_block_size = min(max_block_height, max_block_width)
    block_size = find_largest_power_of_2(max_block_size)

    # Calculate qubits needed
    qubits_needed = 2 * int(math.log2(block_size))

    # Crop the image to fit the blocks perfectly
    cropped_height = blocks_y * block_size
    cropped_width = blocks_x * block_size

    if len(image.shape) == 3:
        cropped_image = image[:cropped_height, :cropped_width, :]
    else:
        cropped_image = image[:cropped_height, :cropped_width]

    # Split the cropped image into blocks and normalize each one
    blocks = []
    normalization_factors = []
    epsilon_stats = {
        "total_zeros_clipped": 0,
        "blocks_with_zeros": 0,
        "epsilon_used": epsilon,
        "per_block_stats": [],
    }

    for y in range(blocks_y):
        for x in range(blocks_x):
            y_start = y * block_size
            y_end = (y + 1) * block_size
            x_start = x * block_size
            x_end = (x + 1) * block_size

            if len(cropped_image.shape) == 3:
                block = cropped_image[y_start:y_end, x_start:x_end, :].astype(
                    np.float32
                )
            else:
                block = cropped_image[y_start:y_end, x_start:x_end].astype(np.float32)

            # Apply epsilon clipping to handle zero values
            block_clipped, block_epsilon_stats = apply_epsilon_clipping(block, epsilon)

            # Track epsilon statistics
            epsilon_stats["total_zeros_clipped"] += block_epsilon_stats["zeros_clipped"]
            if block_epsilon_stats["zeros_clipped"] > 0:
                epsilon_stats["blocks_with_zeros"] += 1
            epsilon_stats["per_block_stats"].append(block_epsilon_stats)

            # Calculate normalization factor for this block (after epsilon clipping)
            block_sum = float(np.sum(block_clipped))

            if normalize_blocks and block_sum > 0.0:
                # Normalize block to sum to 1
                normalized_block = block_clipped / block_sum
                normalization_factors.append(block_sum)
            else:
                # Don't normalize, but still track the sum
                normalized_block = block_clipped
                normalization_factors.append(block_sum if block_sum > 0 else 1.0)

            blocks.append(normalized_block)

    return {
        "blocks": blocks,
        "normalization_factors": normalization_factors,
        "epsilon_info": epsilon_stats,
        "block_size": block_size,
        "qubits_needed": qubits_needed,
        "blocks_y": blocks_y,
        "blocks_x": blocks_x,
        "original_shape": image.shape,
        "cropped_shape": cropped_image.shape,
        "metadata": {
            "num_blocks_requested": num_blocks,
            "num_blocks_created": len(blocks),
            "channels": channels,
            "normalized": normalize_blocks,
            "epsilon_clipped": True,
            "epsilon_value": epsilon,
            "crop_info": {
                "original_height": height,
                "original_width": width,
                "cropped_height": cropped_height,
                "cropped_width": cropped_width,
            },
        },
    }


def reconstruct_from_blocks(
    processed_blocks, split_result, *, denormalize=True, remove_epsilon=False
):
    """
    Reconstruct an image from processed blocks using the original split result metadata

    Parameters:
        processed_blocks (list): List of processed blocks (same order as original split)
        split_result (dict): Result dictionary from split_image_for_quantum
        denormalize (bool): Whether to apply the original normalization factors to restore scale
        remove_epsilon (bool): Whether to attempt to remove epsilon padding (set small values back to 0)

    Returns:
        numpy.ndarray: Reconstructed image
    """
    blocks_y = split_result["blocks_y"]
    blocks_x = split_result["blocks_x"]
    block_size = split_result["block_size"]
    channels = split_result["metadata"]["channels"]
    normalization_factors = split_result["normalization_factors"]
    epsilon = split_result["metadata"]["epsilon_value"]

    # Calculate reconstructed image dimensions
    height = blocks_y * block_size
    width = blocks_x * block_size

    # Create empty image
    if channels == 1:
        reconstructed = np.zeros((height, width), dtype=np.float32)
    else:
        reconstructed = np.zeros((height, width, channels), dtype=np.float32)

    # Place each block back
    block_idx = 0
    for y in range(blocks_y):
        for x in range(blocks_x):
            y_start = y * block_size
            y_end = (y + 1) * block_size
            x_start = x * block_size
            x_end = (x + 1) * block_size

            # Get the processed block
            processed_block = processed_blocks[block_idx].copy()

            # Print diagnostic information
            logger.debug(f"Block {block_idx} before denormalization:")
            logger.debug(f"Sum: {np.sum(processed_block):.6f}")
            logger.debug(f"Min: {np.min(processed_block):.6f}")
            logger.debug(f"Max: {np.max(processed_block):.6f}")

            # Apply denormalization if requested
            if denormalize and split_result["metadata"]["normalized"]:
                processed_block *= normalization_factors[block_idx]
                logger.debug(
                    f"After denormalization (factor: {normalization_factors[block_idx]:.6f}):"
                )
                logger.debug(f"Sum: {np.sum(processed_block):.6f}")
                logger.debug(f"Min: {np.min(processed_block):.6f}")
                logger.debug(f"Max: {np.max(processed_block):.6f}")

            # Remove epsilon padding if requested
            if remove_epsilon:
                epsilon_threshold = epsilon * 2
                processed_block[processed_block <= epsilon_threshold] = 0

            # Place the block
            if channels == 1:
                reconstructed[y_start:y_end, x_start:x_end] = processed_block
            else:
                reconstructed[y_start:y_end, x_start:x_end, :] = processed_block

            block_idx += 1

    return reconstructed



r"""
The "bluequbit.device" device is the base device for Bluequbit Pennylane plugin used to run Pennylane circuits on
BlueQubit platform
"""

# ruff: noqa: ARG002

from __future__ import annotations

import base64
import pickle
import zlib
from collections import defaultdict
from typing import TYPE_CHECKING, ClassVar

from pennylane import numpy as np
from pennylane.devices import QubitDevice
from pennylane.ops.qubit.attributes import diagonal_in_z_basis

import bluequbit.exceptions
from bluequbit import BQClient
from bluequbit.job_metadata_constants import QUEUED_CPU_JOBS_LIMIT

if TYPE_CHECKING:
    from collections.abc import Callable


class BluequbitDevice(QubitDevice):
    """BluequbitDevice device for PennyLane. This device is not directly used but serves as the base class
    for BluequbitCPU and BluequbitGPU subclasses. However, most functionality of the plugin is implemented
    inside this class. The plugin is used to run Pennylane circuits on BlueQubit platform. It requires a
    BlueQubit token, which you can get from your BlueQubit account.

    .. warning::

            To use this plugin, you must have pennylane>=0.39 version installed. It requires
            Python 3.9, but we would recommend using Python 3.10 . Make sure your Python version
            is not older.

    Args:
        wires (int, Iterable[Number, str]): Number of subsystems represented by the device,
            or iterable that contains unique labels for the subsystems as numbers (i.e., ``[-1, 0, 2]``)
            or strings (``['auxiliary', 'q1', 'q2']``). Default 1 if not specified.
        token (str, None): Your BlueQubit token. This is an optional keyword argument that defaults to None.
            If None, the token will be retrieved from the BLUEQUBIT_API_TOKEN environment variable.
    """

    name = "Bluequbit PennyLane plugin"
    short_name = "bluequbit.device"
    pennylane_requires = ">=0.39"
    version = "0.0.1"
    author = "Bluequbit"

    operations: ClassVar[set[str]] = {
        "Identity",
        "Snapshot",
        "BasisState",
        "StatePrep",
        "QubitStateVector",
        "QubitUnitary",
        "ControlledQubitUnitary",
        "MultiControlledX",
        "DiagonalQubitUnitary",
        "SpecialUnitary",
        "PauliX",
        "PauliY",
        "PauliZ",
        "MultiRZ",
        "Hadamard",
        "S",
        "Adjoint(S)",
        "T",
        "Adjoint(T)",
        "SX",
        "Adjoint(SX)",
        "CNOT",
        "SWAP",
        "ISWAP",
        "PSWAP",
        "Adjoint(ISWAP)",
        "SISWAP",
        "Adjoint(SISWAP)",
        "SQISW",
        "CSWAP",
        "Toffoli",
        "CY",
        "CZ",
        "CCZ",
        "CH",
        "PhaseShift",
        "ControlledPhaseShift",
        "CPhase",
        "RX",
        "RY",
        "RZ",
        "Rot",
        "CRX",
        "CRY",
        "CRZ",
        "CRot",
        "IsingXX",
        "IsingYY",
        "IsingZZ",
        "IsingXY",
        "SingleExcitation",
        "SingleExcitationPlus",
        "SingleExcitationMinus",
        "DoubleExcitation",
        "DoubleExcitationPlus",
        "DoubleExcitationMinus",
        "QubitCarry",
        "QubitSum",
        "OrbitalRotation",
        "QFT",
        "ECR",
    }

    observables: ClassVar[set[str]] = {
        "PauliX",
        "PauliY",
        "PauliZ",
        "Hadamard",
        "Hermitian",
        "Identity",
        "Projector",
        "SparseHamiltonian",
        "Hamiltonian",
        "Sum",
        "SProd",
        "Prod",
        "Exp",
    }

    def __init__(self, wires, *args, token=None, **kwargs):
        default_kwargs = {
            "shots": None,
            "analytic": None,
            "r_dtype": np.float64,
            "c_dtype": np.complex128,
        }
        kwargs = {**default_kwargs, **kwargs}

        self._operation_calls = defaultdict(int)
        super().__init__(
            wires,
            shots=kwargs["shots"],
            r_dtype=kwargs["r_dtype"],
            c_dtype=kwargs["c_dtype"],
            analytic=kwargs["analytic"],
        )
        self._debugger = None

        # Create the initial state. The state will always be None.
        self._state = self._create_basis_state(0)
        self._pre_rotated_state = self._state

        self._apply_ops: dict[str, Callable] = {
            "PauliX": self._apply_x,
            "PauliY": self._apply_y,
            "PauliZ": self._apply_z,
            "Hadamard": self._apply_hadamard,
            "S": self._apply_s,
            "T": self._apply_t,
            "SX": self._apply_sx,
            "CNOT": self._apply_cnot,
            "SWAP": self._apply_swap,
            "CZ": self._apply_cz,
            "Toffoli": self._apply_toffoli,
        }

        self._device = "bluequbit.device"
        self._token = token

        try:
            self._bq = BQClient(self._token)
        except bluequbit.exceptions.BQUnauthorizedAccessError:
            raise ValueError(
                "BlueQubit authentication failed. Please provide a valid API token when creating the PennyLane device "
                "using the 'token' parameter (e.g., dev = qml.device('bluequbit.cpu', wires=1, token='<your-token>')) "
                "or set the BLUEQUBIT_API_TOKEN environment variable. "
                "You can find your API token at https://app.bluequbit.io."
            ) from None

    # pylint: disable=arguments-differ
    def apply(self, operations, *args, **kwargs):
        for op in operations:
            self._apply_operation(self._state, op)

    def _apply_operation(self, state, operation):
        self._operation_calls[operation.name] += 1

        if operation.name in self._apply_ops:
            return self._apply_ops[operation.name](state, axes=None)

        wires = operation.wires
        if operation in diagonal_in_z_basis:
            return self._apply_diagonal_unitary(state, None, wires)
        if len(wires) <= 2:
            # Einsum is faster for small gates
            return self._apply_unitary_einsum(state, None, wires)
        return self._apply_unitary(state, None, wires)

    def _apply_x(self, state, axes, **kwargs):
        return [0.0]

    def _apply_y(self, state, axes, **kwargs):
        return [0.0]

    def _apply_z(self, state, axes, **kwargs):
        return [0.0]

    def _apply_hadamard(self, state, axes, **kwargs):
        return [0.0]

    def _apply_s(self, state, axes, inverse=False):
        return [0.0]

    def _apply_t(self, state, axes, inverse=False):
        return [0.0]

    def _apply_sx(self, state, axes, inverse=False):
        return [0.0]

    def _apply_cnot(self, state, axes, **kwargs):
        return [0.0]

    def _apply_swap(self, state, axes, **kwargs):
        return [0.0]

    def _apply_cz(self, state, axes, **kwargs):
        return [0.0]

    def _apply_toffoli(self, state, axes, **kwargs):
        return [0.0]

    def _apply_phase(self, state, axes, parameters, inverse=False):
        return [0.0]

    def expval(self, observable, shot_range=None, bin_size=None):
        return [0.0]

    def var(self, observable, shot_range=None, bin_size=None):
        return [0.0]

    @classmethod
    def capabilities(cls):
        capabilities = super().capabilities().copy()
        capabilities.update(
            model="qubit",
            supports_inverse_operations=True,
            supports_analytic_computation=True,
            supports_broadcasting=False,
            returns_state=True,
            passthru_devices={},
        )
        return capabilities

    @staticmethod
    def _create_basis_state(index):  # noqa: ARG004
        return [0.0]

    @property
    def state(self):
        return [0.0]

    def density_matrix(self, wires):
        return [0.0]

    def _apply_state_vector(self, state, device_wires):
        return [0.0]

    def _apply_basis_state(self, state, wires):
        return [0.0]

    def _apply_unitary(self, state, mat, wires):
        return [0.0]

    def _apply_unitary_einsum(self, state, mat, wires):
        return [0.0]

    def _apply_diagonal_unitary(self, state, phases, wires):
        return [0.0]

    def reset(self):
        self._operation_calls = defaultdict(int)

    def analytic_probability(self, wires=None):
        return [0.0]

    def generate_samples(self):
        """Returns the computational basis samples generated for all wires.
        In the _qubit_device.py, the function calls for analytic_probability for its operations.
        """
        return self.analytic_probability()

    def sample(self, observable, shot_range=None, bin_size=None, counts=False):
        return [0.0]

    def operation_calls(self):
        """Statistics of operation calls"""
        return self._operation_calls

    def execute(self, circuit, **kwargs):
        data = {"circuit": {"pennylane_circuit": circuit}, "method": "execute"}
        result = self._run_bluequbit(data)
        return result

    def batch_execute(self, circuits, **kwargs):
        if type(circuits) is tuple:
            circuits = list(circuits)
        if len(circuits) == 1:
            res = [self.execute(circuits[0])]
        elif len(circuits) > QUEUED_CPU_JOBS_LIMIT:
            raise ValueError(
                f"Cannot run more than {QUEUED_CPU_JOBS_LIMIT} jobs in batch mode"
            )
        else:
            circuit_batch = [{"pennylane_circuit": c} for c in circuits]
            data = {"circuit": circuit_batch, "method": "batch"}
            res = self._run_bluequbit(data)
        return res

    @staticmethod
    def _decode_pennylane_result(result):
        encoded_result = result._pennylane_result  # noqa: SLF001
        decoded_result = base64.b64decode(encoded_result)
        return pickle.loads(zlib.decompress(decoded_result))  # noqa: S301

    def _run_bluequbit(self, data):
        device = data.get("device", self._device)

        result = self._bq.run(
            circuits=data.get("circuit"), device=device, shots=self.shots
        )
        print(result)

        if isinstance(result, list):
            batch_result = [
                self._decode_pennylane_result(single_result) for single_result in result
            ]
            return batch_result
        return self._decode_pennylane_result(result)

    def adjoint_jacobian(self, tape, starting_state=None, use_device_state=False):
        data = {
            "circuit": {"pennylane_circuit": tape},
            "method": "adjoint_jacobian",
            "device": self._device + ".adjoint",
        }
        result = self._run_bluequbit(data)
        return result



from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

from .exceptions import BQSDKUsageError

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


def run_circuit_qiskit(qc: QuantumCircuit, shots: int | None = None):
    from qiskit import transpile
    from qiskit.circuit import Measure
    from qiskit.quantum_info import Statevector
    from qiskit_aer import Aer

    qiskit_backend = Aer.get_backend("aer_simulator")
    transpiled = transpile(qc, qiskit_backend, routing_method="none")
    transpiled.remove_final_measurements()

    has_measurements = any(isinstance(instr, Measure) for instr, _, _ in qc.data)
    output: dict[str, Any] = {
        "has_measurements": has_measurements,
    }
    if has_measurements or (shots is not None and shots > 0):
        if has_measurements and shots is None:
            raise BQSDKUsageError(
                "You have to specify shots when there are measurements."
            )
        run_start = time.time()
        sv = Statevector(transpiled)
        run_end = time.time()
        output["counts"] = sv.probabilities_dict()
    elif shots is None or shots == 0:
        run_start = time.time()
        sv = Statevector(transpiled)
        run_end = time.time()
        output["counts"] = sv.probabilities_dict()
        output["state_vector"] = sv
    else:
        raise BQSDKUsageError(
            f"Wrong configuration for bq.run with device=local: has_measurements={has_measurements}, shots={shots}"
        )
    output["run_start"] = run_start
    output["run_end"] = run_end
    return output


def run_circuit_cirq(circuit, shots: int | None = None):
    simulator: cirq.Simulator | qsimcirq.QSimSimulator
    try:
        # Prefer qsim whenever possible
        import qsimcirq

        simulator = qsimcirq.QSimSimulator()
    except ImportError:
        import cirq

        simulator = cirq.Simulator()

    has_measurements = circuit.has_measurements()
    output = {
        "has_measurements": has_measurements,
    }
    if has_measurements and shots is None:
        raise BQSDKUsageError("You have to specify shots when there are measurements.")
    run_start = time.time()
    result = simulator.simulate(circuit)
    run_end = time.time()
    output["state_vector"] = result.final_state_vector
    output["counts"] = result.final_state_vector * result.final_state_vector.conj()
    output["run_start"] = run_start
    output["run_end"] = run_end
    return output



from __future__ import annotations

from .backend import BlueQubitBackendV2


class BlueQubitProvider:
    """A Qiskit provider for accessing BlueQubit backend.

    :param api_token: API token of the user. If ``None``, the token will be looked
                      in the environment variable BLUEQUBIT_API_TOKEN.
    :param backend: BlueQubit backend to use. If ``None``, a new backend will be created
                    using the API token.
    """

    def __init__(
        self, *, api_token: str | None = None, backend: BlueQubitBackendV2 | None = None
    ):
        super().__init__()
        self.api_token = api_token
        self.backend = backend
        if self.backend is not None:
            assert self.api_token is None

    def get_backend(self, execution_mode: str | None = None):
        """
        :param execution_mode: Execution mode. Can be ``"cloud"``, ``"local"`` or ``None``.
                           If ``None``, the environment variable
                           BLUEQUBIT_EXECUTION_MODE will be used if set, otherwise
                           ``"cloud"`` will be used.
        """
        if self.backend is None:
            self.backend = BlueQubitBackendV2(
                api_token=self.api_token, execution_mode=execution_mode
            )
        return self.backend



import logging

import packaging.version
import requests

logger = logging.getLogger("bluequbit-python-sdk")


def equal_other_than_post_version(pip_version, local_version):
    """Check if the versions are equal except for the post version."""
    return (
        pip_version.major == local_version.major
        and pip_version.minor == local_version.minor
        and pip_version.micro == local_version.micro
        and pip_version.pre == local_version.pre
    )


MESSAGE_TEMPLATE = (
    "There is a {} of BlueQubit Python SDK available on PyPI. We"
    " recommend upgrading. Run 'pip install --upgrade bluequbit' to upgrade"
    " from your version {} to {}."
)


def check_version(version):
    local_version = packaging.version.parse(version)
    if local_version.is_prerelease:
        logger.warning(
            "Beta version %s of BlueQubit Python SDK is being used.", version
        )
    req = requests.get("https://pypi.python.org/pypi/bluequbit/json", timeout=2.0)
    if not req.ok:
        message = "PyPI version check unsuccessful."
        logger.debug(message)
        return message

    # find max version on PyPI
    releases = req.json().get("releases", [])
    pip_version = packaging.version.parse("0")
    for release in releases:
        ver = packaging.version.parse(release)
        if not ver.is_prerelease or local_version.is_prerelease:
            pip_version = max(pip_version, ver)

    if pip_version.major > local_version.major:
        message = MESSAGE_TEMPLATE.format("major upgrade", local_version, pip_version)
        logger.warning(message)
    elif pip_version > local_version and not equal_other_than_post_version(
        pip_version, local_version
    ):
        message = MESSAGE_TEMPLATE.format("newer version", local_version, pip_version)
        logger.info(message)
    else:
        message = ""
    return message



from qiskit import QuantumCircuit
from qiskit.circuit import Parameter


def tfim_2d_grid(
    n_rows: int,
    n_cols: int,
    trotter_steps: int,
) -> QuantumCircuit:
    """
    Create a parametric circuit for the transverse-field Ising model (TFIM) on a 2D
    grid with alternating RX and RZZ layers, representing Trotter evolution steps.

    Args:
        n_rows: Number of rows in the 2D grid.
        n_cols: Number of columns in the 2D grid.
        trotter_steps: Number of Trotter steps (alternating RZZ/RX layer pairs).

    Returns:
        A parametric QuantumCircuit with parameters for RZZ and RX gates.
    """
    n_qubits = n_rows * n_cols
    qc = QuantumCircuit(n_qubits)

    # Build grid connectivity with open boundary conditions
    grid_connectivity = []
    for r in range(n_rows):
        for c in range(n_cols):
            qubit = r * n_cols + c
            if c < n_cols - 1:
                grid_connectivity.append((qubit, r * n_cols + c + 1))
            if r < n_rows - 1:
                grid_connectivity.append((qubit, (r + 1) * n_cols + c))

    # Create Trotter steps with alternating RX and RZZ layers
    for step in range(trotter_steps):
        # RX layer on all qubits
        for i in range(n_qubits):
            param = Parameter(f"θ_rx_{step}_{i}")
            qc.rx(param, i)

        # RZZ layer on grid connectivity
        for i, j in grid_connectivity:
            param = Parameter(f"θ_rzz_{step}_{i}_{j}")
            qc.rzz(param, i, j)

    return qc



# wrapper for QFT to have docstring in the desired style
from __future__ import annotations

from qiskit.circuit.library import QFT as OriginalQFT  # noqa: N811


class QFT(OriginalQFT):
    r"""Quantum Fourier Transform Circuit.

    The Quantum Fourier Transform (QFT) on :math:`n` qubits is the operation

    .. math::

        |j\rangle \mapsto \frac{1}{2^{n/2}} \sum_{k=0}^{2^n - 1} e^{2\pi ijk / 2^n} |k\rangle

    The circuit that implements this transformation can be implemented using Hadamard gates
    on each qubit, a series of controlled-U1 (or Z, depending on the phase) gates and a
    layer of Swap gates. The layer of Swap gates can in principle be dropped if the QFT appears
    at the end of the circuit, since then the re-ordering can be done classically. They
    can be turned off using the ``do_swaps`` attribute.

    :param num_qubits: The number of qubits on which the QFT acts.
    :param approximation_degree: The degree of approximation (0 for no approximation).
    :param do_swaps: Whether to include the final swaps in the QFT.
    :param inverse: If True, the inverse Fourier transform is constructed.
    :param insert_barriers: If True, barriers are inserted as visualization improvement.
    :param name: The name of the circuit.

    For 4 qubits, the circuit that implements this transformation is:

    .. plot::

       from qiskit.circuit.library import QFT
       from qiskit.tools.jupyter.library import _generate_circuit_library_visualization
       circuit = QFT(4)
       _generate_circuit_library_visualization(circuit)

    The inverse QFT can be obtained by calling the ``inverse`` method on this class.
    The respective circuit diagram is:

    .. plot::

       from qiskit.circuit.library import QFT
       from qiskit.tools.jupyter.library import _generate_circuit_library_visualization
       circuit = QFT(4).inverse()
       _generate_circuit_library_visualization(circuit)

    One method to reduce circuit depth is to implement the QFT approximately by ignoring
    controlled-phase rotations where the angle is beneath a threshold. This is discussed
    in more detail in https://arxiv.org/abs/quant-ph/9601018 or
    https://arxiv.org/abs/quant-ph/0403071.

    Here, this can be adjusted using the ``approximation_degree`` attribute: the smallest
    ``approximation_degree`` rotation angles are dropped from the QFT. For instance, a QFT
    on 5 qubits with approximation degree 2 yields (the barriers are dropped in this example):

    .. plot::

       from qiskit.circuit.library import QFT
       from qiskit.tools.jupyter.library import _generate_circuit_library_visualization
       circuit = QFT(5, approximation_degree=2)
       _generate_circuit_library_visualization(circuit)

    """

    def __init__(
        self,
        num_qubits: int | None = None,
        approximation_degree: int = 0,
        do_swaps: bool = True,
        inverse: bool = False,
        insert_barriers: bool = False,
        name: str | None = None,
    ) -> None:
        super().__init__(
            num_qubits, approximation_degree, do_swaps, inverse, insert_barriers, name
        )



import ast
import gzip
import json
import math
from pathlib import Path
from typing import Any

from qiskit import QuantumCircuit, transpile
from qiskit.transpiler import (
    CouplingMap,
    InstructionDurations,
    InstructionProperties,
    Target,
)

# Import BackendV1 for type hinting if available (Qiskit 1.x)
try:
    from qiskit.providers import BackendV1

    HAS_BACKEND_V1 = True
except ImportError:
    HAS_BACKEND_V1 = False


# Mock classes for Qiskit 1.x configuration and properties


class MockGateConfig:
    """Mock GateConfig for Qiskit 1.x BackendV1 compatibility."""

    def __init__(self, name: str, num_qubits: int, parameters: list[str] | None = None):
        self.name = name
        self.num_qubits = num_qubits
        self.parameters = parameters if parameters is not None else []
        self.qasm_def = None  # Not strictly required for transpiler, but common


class MockBackendConfiguration:
    def __init__(self, backend_data: dict):
        self.backend_name = backend_data["backend_name"]
        self.num_qubits = backend_data["num_qubits"]
        self.dt = backend_data.get("dt")
        self.max_shots = backend_data.get("max_shots", 100000)
        self.backend_version = "offline"
        self.simulator = False  # Assume it's a real device config

        # Filter basis gates to only include standard gates
        all_operations = backend_data.get(
            "basis_gates", backend_data["operation_names"]
        )
        standard_operations = {
            "sx",
            "x",
            "rz",
            "ecr",
            "id",
            "measure",
            "reset",
            "delay",
            "cx",
            "cz",
            "h",
            "y",
            "z",
            "s",
            "t",
            "sdg",
            "tdg",
            "rx",
            "ry",
            "u",
            "u1",
            "u2",
            "u3",
            "swap",
            "iswap",
        }
        self.basis_gates = [op for op in all_operations if op in standard_operations]

        # Populate coupling_map
        self.coupling_map = backend_data.get("coupling_map")
        if self.coupling_map and isinstance(self.coupling_map[0], tuple):
            # Convert tuples to lists for JSON compatibility if needed, but here we expect lists
            self.coupling_map = [list(c) for c in self.coupling_map]

        # Generate mock gate configs for the 'gates' attribute
        self.gates = []
        for gate_name in self.basis_gates:
            num_qubits = 1
            parameters: list[str] = []
            if gate_name in ["cx", "cz", "ecr", "swap", "iswap"]:
                num_qubits = 2
            elif gate_name in ["u", "u3"]:
                parameters = ["theta", "phi", "lambda"]
            elif gate_name == "u2":
                parameters = ["phi", "lambda"]
            elif gate_name == "u1":
                parameters = ["lambda"]
            elif gate_name in {"rz", "rx", "ry"}:
                parameters = ["theta"]

            self.gates.append(MockGateConfig(gate_name, num_qubits, parameters))

        # Add "measure" and "reset" if not already in basis_gates, as they are typically present
        if "measure" not in self.basis_gates:
            self.gates.append(MockGateConfig("measure", 1))
        if "reset" not in self.basis_gates:
            self.gates.append(MockGateConfig("reset", 1))


class MockBackendProperties:
    def __init__(self, backend_data: dict):
        self._data = backend_data

    def gate_error(
        self, gate: str, qubits: int | list[int] | tuple[int, ...]
    ) -> float | None:
        qubit_key = str(qubits) if isinstance(qubits, tuple) else str((qubits,))
        return (
            self._data.get("gate_errors", {})
            .get(gate, {})
            .get(qubit_key, {})
            .get("error")
        )

    def gate_length(
        self, gate: str, qubits: int | list[int] | tuple[int, ...]
    ) -> float | None:
        qubit_key = str(qubits) if isinstance(qubits, tuple) else str((qubits,))
        return (
            self._data.get("gate_errors", {})
            .get(gate, {})
            .get(qubit_key, {})
            .get("duration")
        )

    def readout_error(self, qubit: int) -> float | None:
        return (
            self._data.get("gate_errors", {})
            .get("measure", {})
            .get(f"({qubit},)", {})
            .get("error")
        )

    def t1(self, qubit: int) -> float | None:
        return self._data.get("qubit_properties", {}).get(str(qubit), {}).get("t1")

    def t2(self, qubit: int) -> float | None:
        return self._data.get("qubit_properties", {}).get(str(qubit), {}).get("t2")

    def is_qubit_operational(self, qubit: int) -> bool:
        """Check if a qubit is operational.
        Qiskit's ``BackendV2Converter`` requires this method.
        """
        return str(qubit) in self._data.get("qubit_properties", {})

    def qubit_property(self, qubit: int) -> dict | None:
        """Return the properties of a single qubit as a dictionary.

        Qiskit's ``BackendV2Converter`` requires this method and expects a dictionary
        with specific keys (e.g., ``"T1"``, ``"T2"``, etc.) and ``(value, unit)`` tuple values.
        """
        props = self._data.get("qubit_properties", {}).get(str(qubit))

        if props:
            # Return a dictionary formatted for Qiskit's converter
            return {
                "T1": (props.get("t1"), None),  # Map 't1' -> 'T1'
                "T2": (props.get("t2"), None),  # Map 't2' -> 'T2'
                "frequency": (props.get("frequency"), None),
            }
        return None

    def gate_property(self, name: str) -> dict:
        """Return the properties of a gate.

        Qiskit's ``BackendV2Converter`` requires this method.
        It expects a dictionary where keys are qubit tuples and
        values are dictionaries of properties (such as ``"error"`` and ``"duration"``).
        """
        gate_data = self._data.get("gate_errors", {}).get(name, {})
        if not gate_data:
            return {}

        properties_dict = {}
        for qubit_str, props in gate_data.items():
            # Convert string key "(q0,)" or "(q0, q1)" to tuple
            try:
                # Use ast.literal_eval to safely parse the tuple string
                qubit_tuple = ast.literal_eval(qubit_str)
            except (ValueError, SyntaxError):
                continue

            # Format properties for the converter: (value, unit)
            formatted_props = {
                "error": (props.get("error"), None),
                "duration": (props.get("duration"), None),
            }
            properties_dict[qubit_tuple] = formatted_props

        return properties_dict

    def is_gate_operational(self, name: str, qubits: tuple) -> bool:
        """Check if a gate is operational on specific qubits.

        Qiskit's ``BackendV2Converter`` requires this method.
        We define "operational" as "existing in the ``gate_errors`` data".
        """
        # Convert the qubit tuple back to the string key format
        qubit_key = str(qubits)
        gate_data = self._data.get("gate_errors", {}).get(name)

        # Return whether the gate exists and has an entry for this specific qubit_key
        return gate_data and qubit_key in gate_data


class QubitProperties:
    """Mock qubit properties class for offline data."""

    def __init__(self, t1, t2, frequency=None):
        self.t1 = t1
        self.t2 = t2
        self.frequency = frequency


class GateProperties:
    """Mock gate properties class for offline data."""

    def __init__(self, error, duration=None):
        self.error = error
        self.duration = duration


def build_target_from_data(backend_data: dict) -> Target:
    """Build a real Qiskit Target from saved backend data.
    This is much better than mocking all Target methods
    """
    # Create coupling map
    coupling_map = (
        CouplingMap(backend_data["coupling_map"])
        if backend_data.get("coupling_map")
        else None
    )

    # Filter basis gates to only include standard gates
    # Some operations like 'if_else', 'for_loop', 'while_loop', 'switch_case' are control flow, not gates
    all_operations = backend_data.get("basis_gates", backend_data["operation_names"])

    # List of known standard gates and operations
    standard_operations = {
        "sx",
        "x",
        "rz",
        "ecr",
        "id",
        "measure",
        "reset",
        "delay",
        "cx",
        "cz",
        "h",
        "y",
        "z",
        "s",
        "t",
        "sdg",
        "tdg",
        "rx",
        "ry",
        "u",
        "u1",
        "u2",
        "u3",
        "swap",
        "iswap",
    }

    # Filter to only standard operations
    basis_gates = [op for op in all_operations if op in standard_operations]

    # Create a Target using from_configuration
    # This gives us a real Target with all methods/attributes the transpiler needs
    target = Target.from_configuration(
        basis_gates=basis_gates,
        num_qubits=backend_data["num_qubits"],
        coupling_map=coupling_map,
        dt=backend_data.get("dt"),
    )

    # Note: The Target created above has generic properties.
    # We now populate it with detailed gate_errors data for the transpiler to use.
    if "gate_errors" in backend_data:
        for gate_name, gate_data in backend_data["gate_errors"].items():
            # Skip gates not in the target (e.g. non-standard ones we filtered out)
            if gate_name not in target:
                continue

            for qubit_str, props in gate_data.items():
                try:
                    # Parse qubit tuple string, e.g. "(0,)" or "(0, 1)"
                    qubits = ast.literal_eval(qubit_str)
                except (ValueError, SyntaxError):
                    continue

                error = props.get("error")
                duration = props.get("duration")

                if error is not None or duration is not None:
                    target.update_instruction_properties(
                        gate_name,
                        qubits,
                        InstructionProperties(duration=duration, error=error),
                    )

    return target


class GateErrorData:
    """Container for detailed gate error data for fidelity estimation.
    Separate from the Target to avoid interfering with transpilation.
    """

    def __init__(self, backend_data: dict):
        self._data = backend_data

        # Parse qubit properties for fidelity estimation
        self.qubit_properties = {}
        for qubit_str, props in backend_data["qubit_properties"].items():
            qubit = int(qubit_str)
            self.qubit_properties[qubit] = QubitProperties(
                t1=props["t1"], t2=props["t2"], frequency=props.get("frequency")
            )

        # Store operation names for reference
        control_flow_ops = {"if_else", "for_loop", "while_loop", "switch_case"}
        all_ops = backend_data["operation_names"]
        self.operation_names = [op for op in all_ops if op not in control_flow_ops]

    def __getitem__(self, gate_name: str):
        """Get gate data by name - for fidelity estimation."""
        return OfflineGateData(self._data["gate_errors"].get(gate_name, {}))

    def __contains__(self, gate_name: str):
        """Check if gate exists."""
        return gate_name in self._data["gate_errors"]


class OfflineGateData:
    """Mock gate data class for offline target."""

    def __init__(self, gate_data: dict):
        self._data = gate_data

    def get(self, qubits):
        """Get properties for specific qubit(s)."""
        qubit_key = str(qubits) if isinstance(qubits, tuple) else str((qubits,))
        if qubit_key in self._data:
            props = self._data[qubit_key]
            return GateProperties(error=props["error"], duration=props.get("duration"))
        return None

    def keys(self):
        """Get all qubit combinations for this gate."""
        return [ast.literal_eval(k) for k in self._data]


class OfflineBackend:
    """Mock backend class that provides the same interface as an IBM backend
    but uses offline calibration data from a ``.json.gz`` file.

    This class supports both fidelity estimation and offline transpilation.
    """

    def __init__(self, calibration_file: str):
        with gzip.open(calibration_file, "rt") as f:
            self._data = json.load(f)

        self._backend_name_str = self._data[
            "backend_name"
        ]  # Store name as internal string
        self.num_qubits = self._data["num_qubits"]

        # Use real Target for transpilation (no wrapper interference)
        self.target = build_target_from_data(self._data)

        # Store detailed gate error data separately for fidelity estimation
        self._gate_error_data = GateErrorData(self._data)

        # Filter operation_names to exclude control flow operations
        control_flow_ops = {"if_else", "for_loop", "while_loop", "switch_case"}
        all_ops = self._data["operation_names"]
        self.operation_names = [op for op in all_ops if op not in control_flow_ops]

        # Transpilation support
        if self._data.get("coupling_map"):
            self.coupling_map = CouplingMap(self._data["coupling_map"])
        else:
            # Fallback: fully connected (not recommended)
            print(
                f"WARNING: No coupling map found in {calibration_file}\n"
                "Using fully connected topology (may not match real backend)"
            )
            self.coupling_map = CouplingMap.from_full(self.num_qubits)

        # Basis gates for transpilation - use filtered operation_names
        self.basis_gates = self._data.get("basis_gates", self.operation_names)

        # Additional backend properties
        self.dt = self._data.get("dt")
        self.max_shots = self._data.get("max_shots", 100000)

        # Instruction durations (needed by transpiler)
        self.instruction_durations = InstructionDurations()

        # Backend version (for compatibility)
        self.backend_version = "offline"

        # Creation datetime
        self.creation_datetime = self._data.get("creation_datetime")

        # --- Qiskit 1.x compatibility additions ---
        if HAS_BACKEND_V1:
            self._options = None  # Required by BackendV2Converter
            self._configuration = MockBackendConfiguration(self._data)
            self._properties = MockBackendProperties(self._data)
            self._provider = None  # Required by BackendV2Converter

    @property
    def gate_error_data(self):
        """Access to gate error data for fidelity estimation."""
        return self._gate_error_data

    # Required for Qiskit 1.x BackendV2Converter
    def configuration(self):
        return self._configuration

    # Required for Qiskit 1.x BackendV2Converter
    def properties(self):
        return self._properties

    # Required for Qiskit 1.x BackendV2Converter to get backend name
    def name(self):
        return f"OfflineBackend(name={self._backend_name_str})"

    @property
    def provider(self):
        # Required for Qiskit 1.x BackendV2Converter
        return self._provider

    # Minimal BackendV1 interface requirements
    def __init_subclass__(cls, **kwargs):
        if HAS_BACKEND_V1 and not issubclass(cls, BackendV1):
            # This makes OfflineBackend compatible with BackendV1
            # without explicitly inheriting, avoiding MRO issues
            # if BackendV1 is not available.
            cls.__bases__ = (BackendV1, *cls.__bases__)
        super().__init_subclass__(**kwargs)


def load_backend(backend_source: str | Any, service=None) -> Any:
    """Load a backend from either a ``.json.gz`` file or a live IBM backend.

    :param backend_source: Either a path to a ``.json.gz`` file or a backend name string.
    :param service: ``QiskitRuntimeService`` instance (required if ``backend_source`` is a backend name).

    :returns: Backend object (either ``OfflineBackend`` or a live IBM backend).
    :rtype: object
    """

    if isinstance(backend_source, (str, Path)):
        backend_path = Path(backend_source)
        # Check if it's a JSON.GZ file
        if backend_path.suffix == ".gz" and backend_path.exists():
            print(f"Loading offline backend data from {backend_path}")
            return OfflineBackend(str(backend_path))
        # Assume it's a backend name, need service
        if service is None:
            raise ValueError(
                "service parameter required when loading live backend by name"
            )
        print(f"Loading live backend: {backend_source}")
        return service.backend(str(backend_source))
    # Assume it's already a backend object
    return backend_source


def calculate_circuit_fidelity_ibm_fez(circuit, backend, include_readout=True):
    """Estimate the fidelity of a quantum circuit based on backend gate fidelities.

    :param circuit: Transpiled quantum circuit.
    :param backend: IBM backend object (live or offline).
    :param include_readout: Whether to include readout errors in the estimate.

    :returns: A dictionary with fidelity estimates and a detailed breakdown, including:
            - ``fidelity_product``: Product method fidelity estimate (more accurate).
            - ``fidelity_linear``: Linear approximation fidelity estimate.
            - ``total_error``: Sum of all gate errors.
            - ``num_single_qubit_gates``: Count of single-qubit gates.
            - ``num_two_qubit_gates``: Count of two-qubit gates.
            - ``num_measurements``: Count of measurements.
            - ``breakdown``: Detailed breakdown by gate type.
    :rtype: dict
    """

    # For offline backends, use the separate gate error data
    # For live backends, use the target directly
    if hasattr(backend, "gate_error_data"):
        # Offline Backend (Custom GateErrorData structure)
        error_data = backend.gate_error_data
    else:
        # Live Backend (Standard Qiskit Target)
        error_data = backend.target

        # The Qiskit 'Target' object does not natively expose an 'operation_names' list,
        # which is required by the lookup logic below. We inject this attribute here
        # to ensure both Offline and Live objects expose a consistent API.
        if not hasattr(error_data, "operation_names"):
            error_data.operation_names = getattr(backend, "operation_names", [])

    # Track errors for each gate type
    gate_errors = []
    gate_breakdown: dict[str, list[dict[str, Any]]] = {
        "single_qubit": [],
        "two_qubit": [],
        "measurements": [],
    }

    # Analyze each instruction in the circuit
    for instruction_data in circuit.data:
        # Handle both old and new Qiskit API
        if hasattr(instruction_data, "operation"):
            instruction = instruction_data.operation
            qargs = instruction_data.qubits
        else:
            instruction, qargs, _ = instruction_data

        gate_name = instruction.name
        qubit_indices = tuple(circuit.find_bit(q).index for q in qargs)

        # Skip barriers and other non-gate operations
        if gate_name in ["barrier", "delay"]:
            continue

        # Handle measurement separately
        if gate_name == "measure":
            if include_readout:
                for qubit_tuple in [qubit_indices]:
                    gate_props = None

                    # 1. Try looking up using Target.get_instruction_properties (Live/V2 Backend)
                    if hasattr(error_data, "get_instruction_properties"):
                        # Use the dedicated lookup method for the Target object
                        props = error_data.get_instruction_properties(
                            gate_name, qubit_indices
                        )
                        if props is not None:
                            error = (
                                props.error.value if hasattr(props, "error") else None
                            )
                            duration = (
                                props.duration.value
                                if hasattr(props, "duration")
                                else None
                            )
                            if error is not None:
                                # Assuming GateProperties is defined elsewhere
                                gate_props = GateProperties(
                                    error=error, duration=duration
                                )

                    # 2. Fallback: Try looking up in OfflineBackend data structure
                    if (
                        gate_props is None
                        and hasattr(error_data, "__getitem__")
                        and "measure" in error_data
                    ):
                        gate_props = error_data["measure"].get(qubit_tuple)

                    if gate_props is not None and gate_props.error is not None:
                        error = gate_props.error
                        gate_errors.append(error)
                        gate_breakdown["measurements"].append(
                            {
                                "qubit": qubit_tuple[0],
                                "error": error,
                                "fidelity": 1 - error,
                            }
                        )
            continue

        # --- Handle Standard Gates ---
        gate_props = None

        # 1. Try looking up using Target.get_instruction_properties (Live/V2 Backend)
        if hasattr(error_data, "get_instruction_properties"):
            props = error_data.get_instruction_properties(gate_name, qubit_indices)
            if props is not None:
                error = props.error.value if hasattr(props, "error") else None
                duration = props.duration.value if hasattr(props, "duration") else None
                if error is not None:
                    gate_props = GateProperties(error=error, duration=duration)

        # 2. Fallback: Try looking up in OfflineBackend data structure
        elif (
            hasattr(error_data, "operation_names")
            and gate_name in error_data.operation_names
        ):
            gate_props = error_data[gate_name].get(qubit_indices)

        # 3. If found, record the error
        if gate_props is not None and gate_props.error is not None:
            error = gate_props.error
            gate_errors.append(error)

            # Categorize gate
            if len(qubit_indices) == 1:
                gate_breakdown["single_qubit"].append(
                    {
                        "gate": gate_name,
                        "qubit": qubit_indices[0],
                        "error": error,
                        "fidelity": 1 - error,
                    }
                )
            elif len(qubit_indices) == 2:
                gate_breakdown["two_qubit"].append(
                    {
                        "gate": gate_name,
                        "qubits": qubit_indices,
                        "error": error,
                        "fidelity": 1 - error,
                    }
                )

    # Calculate total fidelity using two methods

    # Method 1: Product of fidelities (exact for independent errors)
    total_fidelity_product = 1.0
    for error in gate_errors:
        total_fidelity_product *= 1 - error

    # Method 2: Sum of errors (linear approximation, valid for small errors)
    total_error_sum = sum(gate_errors)
    total_fidelity_linear = 1 - total_error_sum

    # Return comprehensive results
    return {
        "fidelity_product": total_fidelity_product,
        "fidelity_linear": total_fidelity_linear,
        "total_error": total_error_sum,
        "num_single_qubit_gates": len(gate_breakdown["single_qubit"]),
        "num_two_qubit_gates": len(gate_breakdown["two_qubit"]),
        "num_measurements": len(gate_breakdown["measurements"]),
        "breakdown": gate_breakdown,
    }


def get_fidelity_report_string(
    fidelity_result: dict, circuit=None, transpiled_circuit=None
):
    """Get a formatted report of the fidelity estimation as a string.

    :param fidelity_result: Result dictionary from ``estimate_circuit_fidelity()``.
    :param circuit: Original circuit (optional).
    :param transpiled_circuit: Transpiled circuit (optional).
    :returns: A string containing the formatted report.
    :rtype: str
    """

    report_string = ""

    report_string += "=" * 70 + "\n"
    report_string += "CIRCUIT FIDELITY ESTIMATION REPORT\n"
    report_string += "=" * 70 + "\n"

    if circuit is not None:
        report_string += "\nOriginal Circuit:\n"
        report_string += f"  Gates: {len(circuit.data)}\n"
        report_string += f"  Depth: {circuit.depth()}\n"
        report_string += f"  Qubits: {circuit.num_qubits}\n"

    if transpiled_circuit is not None:
        report_string += "\nTranspiled Circuit:\n"
        report_string += f"  Gates: {len(transpiled_circuit.data)}\n"
        report_string += f"  Depth: {transpiled_circuit.depth()}\n"
        report_string += f"  Qubits: {transpiled_circuit.num_qubits}\n"

    report_string += "\n" + "=" * 70 + "\n"
    report_string += "FIDELITY ESTIMATES:\n"
    report_string += "=" * 70 + "\n"
    report_string += f"\nEstimated circuit fidelity (product method): {fidelity_result['fidelity_product']:.6f}\n"
    report_string += f"Estimated circuit fidelity (linear approx):  {fidelity_result['fidelity_linear']:.6f}\n"
    report_string += f"Total error sum: {fidelity_result['total_error']:.6f}\n"

    report_string += "\nGate counts:\n"
    report_string += (
        f"  Single-qubit gates: {fidelity_result['num_single_qubit_gates']}\n"
    )
    report_string += f"  Two-qubit gates:    {fidelity_result['num_two_qubit_gates']}\n"
    report_string += f"  Measurements:       {fidelity_result['num_measurements']}\n"

    report_string += "\n" + "=" * 70 + "\n"
    report_string += "DETAILED GATE BREAKDOWN:\n"

    report_string += "=" * 70 + "\n"

    report_string += "\nSingle-Qubit Gates (first 10):\n"
    for i, gate in enumerate(fidelity_result["breakdown"]["single_qubit"][:10], 1):
        report_string += f"  {i:2d}. {gate['gate'].upper():4s} on qubit {gate['qubit']:3d} - error: {gate['error']:.6f}, fidelity: {gate['fidelity']:.6f}\n"
    if fidelity_result["num_single_qubit_gates"] > 10:
        report_string += (
            f"  ... and {fidelity_result['num_single_qubit_gates'] - 10} more\n"
        )

    report_string += "\nTwo-Qubit Gates (first 10):\n"
    for i, gate in enumerate(fidelity_result["breakdown"]["two_qubit"][:10], 1):
        report_string += (
            f"  {i:2d}. {gate['gate'].upper():4s} on qubits {gate['qubits']} - "
        )
        report_string += (
            f"error: {gate['error']:.6f}, fidelity: {gate['fidelity']:.6f}\n"
        )
    if fidelity_result["num_two_qubit_gates"] > 10:
        report_string += (
            f"  ... and {fidelity_result['num_two_qubit_gates'] - 10} more\n"
        )

    if fidelity_result["num_measurements"] > 0:
        report_string += "\nMeasurements:\n"
        for i, gate in enumerate(fidelity_result["breakdown"]["measurements"], 1):
            report_string += f"  {i:2d}. MEAS on qubit {gate['qubit']:3d} - "
            report_string += (
                f"error: {gate['error']:.6f}, fidelity: {gate['fidelity']:.6f}\n"
            )

    # Calculate error contribution by gate type for the product method (Logarithmic weight)
    f_single = 1.0
    for g in fidelity_result["breakdown"]["single_qubit"]:
        f_single *= 1.0 - g["error"]

    f_two = 1.0
    for g in fidelity_result["breakdown"]["two_qubit"]:
        f_two *= 1.0 - g["error"]

    f_meas = 1.0
    for g in fidelity_result["breakdown"]["measurements"]:
        f_meas *= 1.0 - g["error"]

    f_total = fidelity_result["fidelity_product"]

    if 0.0 < f_total < 1.0:
        log_total = math.log(f_total)
        share_single = (math.log(f_single) / log_total) * 100
        share_two = (math.log(f_two) / log_total) * 100
        share_meas = (math.log(f_meas) / log_total) * 100
    else:
        share_single = 0.0
        share_two = 0.0
        share_meas = 0.0

    report_string += "\n" + "=" * 70 + "\n"
    report_string += "ERROR CONTRIBUTION BY GATE TYPE (PRODUCT METHOD):\n"
    report_string += "=" * 70 + "\n"
    report_string += "This shows what the fidelity would be if ONLY that specific error type existed,\n"
    report_string += "and its relative contribution to the total fidelity decay.\n"
    report_string += "-" * 70 + "\n"
    report_string += f"\nSingle-qubit component:  {f_single:.6f} (Contributes {share_single:.1f}% to decay)\n"
    report_string += f"Two-qubit component:     {f_two:.6f} (Contributes {share_two:.1f}% to decay)\n"
    report_string += f"Measurement component:   {f_meas:.6f} (Contributes {share_meas:.1f}% to decay)\n"
    report_string += f"{'─' * 70}\n"
    report_string += f"Combined Product:        {f_total:.6f}\n"

    # Calculate error contribution by gate type for the linear approximation method
    single_error = sum(g["error"] for g in fidelity_result["breakdown"]["single_qubit"])
    two_error = sum(g["error"] for g in fidelity_result["breakdown"]["two_qubit"])
    meas_error = sum(g["error"] for g in fidelity_result["breakdown"]["measurements"])

    if fidelity_result["total_error"] > 0:
        report_string += "\n" + "=" * 70 + "\n"
        report_string += (
            "ERROR CONTRIBUTION BY GATE TYPE (LINEAR APPROXIMATION METHOD):\n"
        )
        report_string += "=" * 70 + "\n"
        report_string += f"\nSingle-qubit gates:  {single_error:.6f} ({single_error / fidelity_result['total_error'] * 100:.1f}%)\n"
        report_string += f"Two-qubit gates:     {two_error:.6f} ({two_error / fidelity_result['total_error'] * 100:.1f}%)\n"
        report_string += f"Measurements:        {meas_error:.6f} ({meas_error / fidelity_result['total_error'] * 100:.1f}%)\n"
        report_string += f"{'─' * 70}\n"
        report_string += (
            f"Total error sum:         {fidelity_result['total_error']:.6f} (100.0%)\n"
        )

    report_string += "\n" + "=" * 70 + "\n"
    report_string += "NOTES:\n"
    report_string += "=" * 70 + "\n"
    report_string += """
• The 'product method' is more accurate (multiplies 1-error for each gate)
• The 'linear approximation' simply sums errors (valid for small errors)
• For error rates < 5%, both methods give similar results
• This estimate assumes:
    - Gate errors are independent
    - No crosstalk between qubits
    - Coherence times >> gate durations
    - No state preparation errors
"""
    return report_string


def estimate_circuit_fidelity_ibm_fez(
    circuit: QuantumCircuit, ibm_token: str = "", verbose: bool = False
) -> float:
    """Estimate the fidelity of a quantum circuit on the IBM FEZ backend.

    :param circuit: The quantum circuit to estimate the fidelity of.
    :param ibm_token: The IBM Quantum token to use for the online mode.
    :param verbose: Whether to print verbose output.

    :returns: The estimated fidelity of the circuit as a float.
    :rtype: float
    """

    # Configuration
    ibm_instance = "crn:v1:bluemix:public:quantum-computing:us-east:a/373146ee8c834aa5a059a09701d8d31b:fd7f08b1-6d70-442e-bf8b-6df950815350::"
    backend_name = "ibm_fez"

    # Get the directory containing this module and construct the calibration file path
    module_dir = Path(__file__).parent
    offline_data_file = module_dir / "ibm_fez_calibration.json.gz"
    if not offline_data_file.exists():
        raise FileNotFoundError(f"Offline data file not found: {offline_data_file}")

    # Determine whether to use online or offline mode
    use_offline = not ibm_token

    if use_offline:
        if verbose:
            mode_info_string = ""
            mode_info_string += "=" * 70 + "\n"
            mode_info_string += "OFFLINE MODE: Using saved calibration data\n"
            mode_info_string += "=" * 70
            print(mode_info_string)
        backend = load_backend(offline_data_file)
        if verbose:
            print(f"Calibration data retrieved at: {backend.creation_datetime}")
    else:
        if verbose:
            mode_info_string = ""
            mode_info_string += "=" * 70 + "\n"
            mode_info_string += "ONLINE MODE: Connecting to IBM Quantum\n"
            mode_info_string += "=" * 70
            print(mode_info_string)

        try:
            from qiskit_ibm_runtime import QiskitRuntimeService
        except ImportError:
            raise ImportError(
                "qiskit_ibm_runtime is not installed. Please install it with `pip install qiskit-ibm-runtime`."
            ) from None

        service = QiskitRuntimeService(
            channel="ibm_cloud", token=ibm_token, instance=ibm_instance
        )
        backend = load_backend(backend_name, service)

    if verbose:
        # Handle backend.name being a method (Qiskit V1/OfflineBackend) or property (Qiskit V2)
        backend_info_string = ""
        b_name = backend.name() if callable(backend.name) else backend.name
        backend_info_string += f"Using backend: {b_name}\n\n"
        backend_info_string += f"\nThe circuit has {len(circuit.data)} gates\n"
        backend_info_string += f"\nTranspiling for {b_name}..."
        print(backend_info_string)

    # Transpile circuit (works both online and offline)
    tqc = transpile(circuit, backend=backend, optimization_level=3)
    if verbose:
        transpilation_info_string = ""
        transpilation_info_string += (
            f"Transpiled circuit has {len(tqc.data)} operations\n"
        )
        if use_offline:
            transpilation_info_string += (
                "✓ Transpilation completed using saved coupling map (offline mode)"
            )
        print(transpilation_info_string)

    # Estimate fidelity
    if verbose:
        print("\nEstimating circuit fidelity...\n")
    fidelity_result = calculate_circuit_fidelity_ibm_fez(
        tqc, backend, include_readout=True
    )

    # Print detailed report
    if verbose:
        report_string = get_fidelity_report_string(
            fidelity_result, circuit=circuit, transpiled_circuit=tqc
        )
        print(report_string)

    return fidelity_result["fidelity_product"]



from __future__ import annotations

import base64
import datetime
import json
import zlib
from io import BytesIO

import numpy as np
import packaging
from dateutil.parser import parse

from . import job_metadata_constants
from .exceptions import (
    BQJobHasNotFinishedYetError,
    BQJobNotCompleteError,
    BQJobStatevectorNotAvailableError,
    BQSDKUsageError,
)
from .http_utils import request_retriable
from .version import __version__

SDK_VERSION = packaging.version.parse(__version__)


class JobResult:
    """This class contains information from a job run on quantum hardware or
    quantum simulators. Mainly it contains the resulting statevector from the
    run. It might contain only partial information, such as job ID,
    when :func:`BQClient.run` is called with ``asynchronous=True``."""

    def __init__(self, data, _backend_connection=None):
        super().__init__()

        self._backend_connection = _backend_connection

        #: str: job id
        self.job_id = data.get("job_id")

        #: str: batch id
        self.batch_id = data.get("batch_id")

        #: str: job run status, can be one of ``"FAILED_VALIDATION"`` | ``"PENDING"`` |
        #       ``"QUEUED"`` | ``"RUNNING"`` | ``"TERMINATED"`` | ``"CANCELED"`` |
        #       ``"NOT_ENOUGH_FUNDS"`` | ``"COMPLETED"``
        self.run_status = data.get("run_status")

        #: str: job name
        self.job_name = data.get("job_name")

        #: str: run device
        self.device = data.get("device")

        #: dict: device parameters used in circuit execution
        self.device_options = data.get("device_options")

        #: int: estimated runtime in milliseconds
        self.estimated_runtime = data.get("estimated_runtime_ms")

        #: float: estimated cost in US dollars
        self.estimated_cost = data.get("estimated_cost")
        if self.estimated_cost is not None:
            self.estimated_cost /= 100.0

        #: datetime: job creation date in UTC timezone
        self.created_on = parse(data.get("created_on"))

        self._results_path = data.get("results_path")

        self._top_100k_results = None

        #: dict: top 128 results
        self.top_128_results = data.get("top_128")

        self._pennylane_result = data.get("pennylane_result")

        #: int: number of qubits
        self.num_qubits = data.get("num_qubits")

        #: int: number of used qubits
        self.num_qubits_used = data.get("num_qubits_used", self.num_qubits)

        #: dict[str, Any]: tags, a JSON serializable dictionary.
        self.tags = data.get("tags")

        #: list[tuple[str, float]]: expectation value
        self.pauli_sum = data.get("pauli_sum")

        #: float | list[float]: expectation value
        self.expectation_value = data.get("expectation_value")

        self.run_start = data.get("run_start")
        if self.run_start is not None:
            self.run_start = parse(self.run_start)

        self.run_end = data.get("run_end")
        if self.run_end is not None:
            self.run_end = parse(self.run_end)

        #: int: job queue time in milliseconds
        self.queue_time_ms = None
        if self.run_start is not None:
            self.queue_time_ms = int(
                (self.run_start - self.created_on) / datetime.timedelta(milliseconds=1)
            )

        #: int: job runtime in milliseconds
        self.run_time_ms = None
        if self.run_start is not None and self.run_end is not None:
            self.run_time_ms = round(
                (self.run_end - self.run_start) / datetime.timedelta(milliseconds=1)
            )

        #: str: error message if failed
        self.error_message = data.get("error_message")

        #: float: job cost in US dollars
        self.cost = data.get("cost")
        if self.cost is not None:
            self.cost /= 100.0

        #: bool: if statevector available
        self.has_statevector = data.get("has_statevector")

        # This is present by default only on results by _run_circuit_local.
        self._statevector = data.get("state_vector")

        #: Cirq, Qiskit, dict: original circuit
        self.circuit = data.get("qc_unprocessed")
        if self.circuit is not None:
            self.circuit = json.loads(
                zlib.decompress(base64.b64decode(self.circuit)).decode()
            )

        self.sdk_version = data.get("sdk_version")
        if self.sdk_version is not None:
            self.sdk_version = packaging.version.parse(self.sdk_version)

        #: int: number of shots specified during the job submission
        self.shots = data.get("shots")

        #: dict | None: additional results from the job execution (e.g., mps_build_time)
        self.run_results = data.get("run_results")

    def get_statevector(self) -> np.ndarray:
        """Return statevector of the job. If the statevector is too large then throws exception.

        :rtype: NumPy array
        """
        if self.run_status in job_metadata_constants.JOB_NON_TERMINAL_STATES:
            raise BQJobHasNotFinishedYetError(self.job_id, self.run_status)
        if not self.ok:
            raise BQJobNotCompleteError(
                self.job_id, self.run_status, self.error_message
            )
        if self.has_statevector is not True:
            raise BQJobStatevectorNotAvailableError(
                self.job_id, self.num_qubits_used, self.shots, self.expectation_value
            )

        if self._statevector is None:
            response = request_retriable("GET", self._results_path + "statevector.txt")
            self._statevector = np.loadtxt(
                BytesIO(response.content), dtype=np.complex128
            )

        return self._statevector

    def get_counts(self) -> None | dict[str, float]:
        """
        If shots argument in bq.run() was not None, return a number of samples
        from the measurements, otherwise, measurement probabilities of the
        computation result, without any statistical fluctuation are returned. In both
        cases only top 131072 (2^17) results are returned.
        """
        if self.top_128_results is not None and len(self.top_128_results) < 128:
            return self.top_128_results
        if self.run_status in job_metadata_constants.JOB_NON_TERMINAL_STATES:
            raise BQJobHasNotFinishedYetError(self.job_id, self.run_status)
        if not self.ok:
            raise BQJobNotCompleteError(
                self.job_id, self.run_status, self.error_message
            )
        if self._top_100k_results is None:
            response = request_retriable("GET", self._results_path + "metadata.json")
            # in case of old jobs that have hybrid_code_result as not None we didn't upload metadata.json
            if response.ok:
                self._top_100k_results = json.loads(response.content).get("counts")
            else:
                raise BQSDKUsageError("Counts are not available for this job")

        return self._top_100k_results

    @property
    def ok(self) -> bool:  # pylint: disable=invalid-name
        """``True``, if job's current run status is ``"COMPLETED"``,
        else ``False``."""

        return self.run_status == "COMPLETED"

    def __repr__(self):
        jr_repr = f"Job ID: {self.job_id}"
        if self.batch_id is not None:
            jr_repr += f" (batch ID: {self.batch_id})"
        if self.job_name is not None:
            jr_repr += f", name: {self.job_name}"
        jr_repr += f", device: {self.device}"
        jr_repr += f", run status: {self.run_status}"
        jr_repr += (
            f", created on: {self.created_on.replace(microsecond=0, tzinfo=None)} UTC"
        )
        if (
            self.run_status in ["PENDING", "QUEUED", "RUNNING"]
            and self.estimated_runtime is not None
        ):
            jr_repr += f", estimated runtime: {self.estimated_runtime} ms"
            jr_repr += f", estimated cost: ${self.estimated_cost:.2f}"
        if self.cost is not None:
            jr_repr += f", cost: ${self.cost:.2f}"
        if self.run_time_ms is not None:
            jr_repr += f", run time: {self.run_time_ms} ms"
        if self.queue_time_ms is not None:
            jr_repr += f", queue time: {self.queue_time_ms} ms"
        if self.num_qubits is not None:
            jr_repr += f", num qubits: {self.num_qubits}"
        if self.num_qubits_used != self.num_qubits:
            jr_repr += f", num qubits used: {self.num_qubits_used}"
        if self.shots is not None:
            jr_repr += f", shots: {self.shots}"
        if self.error_message is not None:
            jr_repr += f", error_message: {self.error_message}"
        return jr_repr



from . import job_metadata_constants

AUTH_ERROR_MESSAGE = (
    "BlueQubit client was not authorized. For full functionality please specify an API token to init(<your-api-token>) or have API"
    " token set in BLUEQUBIT_API_TOKEN environment variable. You can"
    " find your API token once you log in to https://app.bluequbit.io"
)


class BQError(Exception):
    def __init__(self, message="Unknown Error Message"):
        super().__init__(message)
        self.message = message


class BQSDKUsageError(BQError):
    def __init__(self, message: str):
        super().__init__(message)


class BQUnauthorizedAccessError(BQError):
    def __init__(self, request_id):
        super().__init__(AUTH_ERROR_MESSAGE + f". Request ID: {request_id}")
        self.request_id = request_id


class BQJobStatevectorNotAvailableError(BQError):
    def __init__(self, job_id, num_qubits, shots, pauli_sum):
        message = "Statevector is not available."
        if shots is not None and shots != 0:
            message += " Job run with shots > 0. Please use .get_counts() instead."
        elif num_qubits > job_metadata_constants.MAX_QUBITS_WITH_STATEVEC:
            message += f" Job has too many, {num_qubits}, qubits."
        elif pauli_sum is not None:
            message += " Observables are provided."
        super().__init__(message)
        self.job_id = job_id
        self.num_qubits = num_qubits


class BQJobNotCompleteError(BQError):
    def __init__(self, job_id, run_status, error_message):
        message = f"Job {job_id} finished with status: {run_status}. {error_message}"
        if run_status == "NOT_ENOUGH_FUNDS":
            message += " Please check out our subscription plans at https://app.bluequbit.io/billing to upgrade your account."
        super().__init__(
            message,
        )
        self.job_id = job_id
        self.run_status = run_status
        self.error_message = error_message


class BQJobHasNotFinishedYetError(BQError):
    def __init__(self, job_id, run_status):
        super().__init__(
            f"Job {job_id} has not finished yet. Current status: {run_status}.",
        )
        self.job_id = job_id
        self.run_status = run_status


class BQJobCouldNotCancelError(BQError):
    def __init__(self, job_id, run_status, error_message):
        super().__init__(
            f"Couldn't cancel job {job_id}. Finished status is {run_status}."
            f" {error_message}",
        )
        self.job_id = job_id
        self.run_status = run_status


class BQBatchJobsLimitExceededError(BQError):
    def __init__(self, num):
        super().__init__(
            "Batch operations job count limit exceeded. The maximum is"
            f" {job_metadata_constants.MAXIMUM_NUMBER_OF_BATCH_JOBS}, but {num} was"
            " used."
        )
        self.num = num


class BQJobsLimitExceededError(BQError):
    def __init__(self, num):
        super().__init__(
            "Maximum number of batch jobs to run is "
            f" {job_metadata_constants.MAXIMUM_NUMBER_OF_JOBS_FOR_RUN}, but {num} was"
            " used."
        )
        self.num = num


class BQCPUJobsLimitExceededError(BQError):
    def __init__(self, num: int):
        super().__init__(
            "Maximum free CPU jobs limit is "
            f" {job_metadata_constants.QUEUED_CPU_JOBS_LIMIT}, but {num} was"
            " used."
        )
        self.num = num


class BQAPIError(BQError):
    def __init__(self, http_status_code, error_message, request_id):
        super().__init__(
            f"{error_message.strip()}. HTTP response status code: {http_status_code}. Request ID: {request_id}",
        )
        self.http_status_code = http_status_code
        self.error_message = error_message
        self.request_id = request_id


class BQJobInvalidDeviceTypeError(BQError):
    def __init__(self, device):
        super().__init__(
            f"Invalid device type {device}. Must be one of"
            f" {', '.join(job_metadata_constants.DEVICE_TYPES)}."
        )


class BQJobExceedsSerializedCircuitSizeLimitError(BQError):
    def __init__(self, size):
        super().__init__(
            f"Circuit size is {size} bytes which exceeds {job_metadata_constants.MAXIMUM_SERIALIZED_CIRCUIT_SIZE / 1_000_000:.0f} MB limit."
        )


class BQJobsMalformedShotsError(BQError):
    def __init__(self, shots, device, max_shots):
        super().__init__(
            f"Malformed number of shots: {shots} or exceeds limit of"
            f" {max_shots} for {device} device."
        )



"""API functions for interacting with peaked circuits on the BlueQubit backend."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from bluequbit.exceptions import BQSDKUsageError

if TYPE_CHECKING:
    from bluequbit.backend_connection import BackendConnection


def get_peaked_circuit(
    _connection: BackendConnection, difficulty: int, circuit_id: int | None = None
) -> dict[str, Any]:
    """Fetch a single peaked circuit by ID from the backend.

    :param _connection: backend connection instance
    :param difficulty: difficulty of the circuit to fetch
    :param circuit_id: non-negative integer ID of the circuit to fetch. If ``None``,
                       a random circuit of the given difficulty will be returned.
    :return: dict with keys ``qasm_str``, ``key``, and ``metadata``
    :raises BQSDKUsageError: if circuit_id is not a non-negative integer
    """
    if circuit_id is not None and (not isinstance(circuit_id, int) or circuit_id < 0):
        raise BQSDKUsageError("circuit_id must be a non-negative integer")
    if difficulty is None or not isinstance(difficulty, int) or difficulty <= 0:
        raise BQSDKUsageError("difficulty must be a positive integer")
    if circuit_id is None:
        params = {"difficulty": difficulty}
    else:
        params = {"difficulty": difficulty, "id": circuit_id}
    response = _connection.send_request(
        req_type="GET", path="/peaked-circuits/file", params=params
    )
    return json.loads(response.text)



from __future__ import annotations

from pathlib import Path

from setuptools import setup
from setuptools.command.develop import develop

try:
    from setuptools.command.editable_wheel import editable_wheel
except ImportError:  # pragma: no cover - setuptools always provides this in dev env
    editable_wheel = None


def exclude_lines_between_markers(
    file_path: str,
    start_marker: str = "<!-- EXCLUDE -->",
    end_marker: str = "<!-- /EXCLUDE -->",
) -> str:
    """
    Reads a file and returns its content with lines between the given markers excluded.
    """
    output_lines = []
    exclude = False
    with Path(file_path).open(encoding="utf-8") as f:
        for line in f:
            if start_marker in line and not exclude:
                exclude = True
                continue
            if end_marker in line and exclude:
                exclude = False
                continue
            if not exclude:
                output_lines.append(line)
    return "".join(output_lines)


readme = exclude_lines_between_markers("README.md")

with Path("src/bluequbit/version.py").open() as f:
    Version = f.read()

Version = Version.rstrip()
Version = Version[15:-1]


def _add_source_only_console_script(distribution) -> None:
    """Register console scripts that should exist only for source installs."""
    entry_points = distribution.entry_points or {}
    console_scripts = list(entry_points.get("console_scripts", []))
    block_image_loader_script = (
        "bluequbit_block_image_loader="
        "bluequbit.library.hierarchical_circuit.block_image_loader_cli:main"
    )
    pps_image_classification_script = (
        "bluequbit_pps_image_classification="
        "bluequbit.library.pps_image_classification.cli:main"
    )
    if block_image_loader_script not in console_scripts:
        console_scripts.append(block_image_loader_script)
    if pps_image_classification_script not in console_scripts:
        console_scripts.append(pps_image_classification_script)
    entry_points["console_scripts"] = console_scripts
    distribution.entry_points = entry_points


class SourceDevelopCommand(develop):
    """Install source-only console scripts during editable/develop installs."""

    def run(self):
        _add_source_only_console_script(self.distribution)
        super().run()


cmdclass = {"develop": SourceDevelopCommand}

if editable_wheel is not None:

    class SourceEditableWheelCommand(editable_wheel):
        """Install source-only console scripts into editable wheels only."""

        def run(self):
            _add_source_only_console_script(self.distribution)
            super().run()

    cmdclass["editable_wheel"] = SourceEditableWheelCommand

setup(
    version=Version,
    long_description=readme,
    long_description_content_type="text/markdown",
    cmdclass=cmdclass,
)
